import pytest

from illumio.policyobjects import (
    Label,
    LabelSet,
    Service,
    ServicePort
)
from illumio.rules import RuleSet
from illumio.workloads import Workload

from helpers import random_string


@pytest.fixture
def role_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='role',
            value='{}-R-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)


@pytest.fixture
def app_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='app',
            value='{}-A-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)


@pytest.fixture
def env_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='env',
            value='{}-E-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)


@pytest.fixture
def loc_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='loc',
            value='{}-L-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)


@pytest.fixture
def web_service(pce, session_identifier):
    identifier = random_string()
    service = pce.services.create(
        Service(
            name='{}-S-WEB-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            service_ports=[
                ServicePort(port=80, proto='tcp'),
                ServicePort(port=443, proto='tcp')
            ],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield service
    pce.services.delete(service.href)


@pytest.fixture
def well_known_service(pce, session_identifier):
    identifier = random_string()
    service = pce.services.create(
        Service(
            name='{}-S-WELL-KNOWN-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            service_ports=[
                ServicePort(port=1, to_port=1023, proto='tcp')
            ],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield service
    pce.services.delete(service.href)


@pytest.fixture
def rdp_service(pce, session_identifier):
    identifier = random_string()
    service = pce.services.create(
        Service(
            name='{}-S-RDP-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            service_ports=[
                ServicePort(port=3389, proto='tcp'),
                ServicePort(port=3389, proto='udp')
            ],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield service
    pce.services.delete(service.href)


@pytest.fixture
def rule_set(pce, session_identifier, app_label, env_label, loc_label):
    identifier = random_string()
    rule_set = pce.rule_sets.create(
        RuleSet(
            name='{}-IPL-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            enabled=True,
            scopes=[LabelSet(labels=[app_label, env_label, loc_label])],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield rule_set
    pce.rule_sets.delete(rule_set.href)


@pytest.fixture
def workload(pce, session_identifier):
    identifier = random_string()
    hostname = '{}.{}'.format(session_identifier, identifier)
    workload = pce.workloads.create(
        Workload(
            name=hostname,
            hostname=hostname,
            description='Created by illumio python library integration tests',
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield workload
    pce.workloads.delete(workload.href)



# -*- coding: utf-8 -*-

"""This module provides a helper for actor objects in rules and enforcement boundaries.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import re
from dataclasses import dataclass

from illumio.exceptions import IllumioException
from illumio.util import href_from, JsonObject, Reference, HREF_REGEX

AMS = 'ams'


@dataclass
class Actor(JsonObject):
    actors: str = None
    label: Reference = None
    label_group: Reference = None
    workload: Reference = None
    virtual_service: Reference = None
    virtual_server: Reference = None
    ip_list: Reference = None

    @staticmethod
    def from_reference(reference):
        href = href_from(reference)
        if href.lower() == AMS:  # special case for all objects
            return Actor(actors=AMS)
        actor = Actor()
        match = re.match(HREF_REGEX, href)
        if match:
            # HREF object types are plural, so we remove the s
            object_type = match.group('type')[:-1]
            setattr(actor, object_type, Reference(href=href))
        else:
            raise IllumioException('Invalid HREF in policy provision changeset: {}'.format(href))
        return actor



# -*- coding: utf-8 -*-

"""This module provides classes related to labels and label groups.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import json
from dataclasses import dataclass
from typing import List

from illumio.util import JsonObject, Reference, MutableObject, pce_api


@dataclass
class LabelUsage(JsonObject):
    label_group: bool = None
    ruleset: bool = None
    rule: bool = None
    static_policy_scopes: bool = None
    containers_inherit_host_policy_scopes: bool = None
    blocked_connection_reject_scope: bool = None
    enforcement_boundary: bool = None


@dataclass
@pce_api('labels')
class Label(MutableObject):
    """Represents a label in the PCE.

    Labels help to configure the reach of policy rules in a dynamic way,
    without relying on precise identifiers like IP addresses.

    When fetching Labels from the PCE, a breakdown of the labels' usage can be
    optionally included.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/labels-and-label-groups.htm

    Usage:
        >>> from illumio import PolicyComputeEngine, Label
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> label = Label(key='role', value='R-DB')
        >>> label = pce.labels.create(label)
        >>> label
        Label(
            href='/orgs/1/labels/18',
            key='role',
            value='R-DB',
            ...
        )
    """
    key: str = None
    value: str = None
    deleted: bool = None
    usage: LabelUsage = None


@dataclass
@pce_api('label_groups', is_sec_policy=True)
class LabelGroup(Label):
    """Represents a label group in the PCE.

    Label groups can contain labels and other sub-groups to define broader
    categories that are often grouped when writing rules or otherwise
    referencing multiple labels.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/labels-and-label-groups.htm

    Usage:
        >>> from illumio import PolicyComputeEngine, LabelGroup
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> pce.set_credentials('api_key_username', 'api_key_secret')
        >>> dev_label = pce.labels.create({'key': 'env', 'value': 'E-DEV'})
        >>> stage_label = pce.labels.create({'key': 'env', 'value': 'E-STAGE'})
        >>> label_group = LabelGroup(
        ...     key='role',
        ...     name='LG-E-PREPROD',
        ...     labels=[dev_label, stage_label]
        ... )
        >>> label_group = pce.label_groups.create(label_group)
        >>> label_group
        LabelGroup(
            href='/orgs/1/sec_policy/draft/label_groups/5704a6f4-e051-4f88-9149-713ee22b5d41',
            key='role',
            value='R-DB',
            ...
        )
    """
    labels: List[Reference] = None
    sub_groups: List['LabelGroup'] = None


@dataclass
class LabelSet(JsonObject):
    labels: List[Reference] = None

    def _encode(self):
        json_array = []
        for label in self.labels:
            key = 'label_group' if '/label_groups/' in label.href else 'label'
            json_array.append({key: label.to_json()})
        return json_array

    @classmethod
    def from_json(cls, data) -> 'LabelSet':
        data = json.loads(data) if type(data) is str else data
        labels = []
        for label_entry in data:
            key = 'label' if 'label' in label_entry else 'label_group'
            labels.append(Label.from_json(label_entry[key]))
        return LabelSet(labels=labels)



# -*- coding: utf-8 -*-

"""
Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .containercluster import *
from .network import *
from .secureconnectgateway import *



import json
import os
import re
from typing import List

import pytest

from illumio.policyobjects import Service, ServicePort
from illumio.workloads import Workload

MOCK_WORKLOADS = os.path.join(pytest.DATA_DIR, 'workloads.json')


@pytest.fixture(scope='module')
def workloads() -> List[dict]:
    with open(MOCK_WORKLOADS, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(scope='module')
def new_workload() -> Workload:
    return Workload(
        name='db0.internal.labs.io',
        hostname='db0.internal.labs.io'
    )


@pytest.fixture(autouse=True)
def workloads_mock(pce_object_mock, workloads):
    pce_object_mock.add_mock_objects(workloads)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    profile_pattern = re.compile('/workloads')
    requests_mock.register_uri('GET', profile_pattern, json=get_callback)
    requests_mock.register_uri('POST', profile_pattern, json=post_callback)
    requests_mock.register_uri('PUT', profile_pattern, json=put_callback)
    requests_mock.register_uri('DELETE', profile_pattern, json=delete_callback)


@pytest.fixture()
def mock_workload(pce):
    yield pce.workloads.get_by_reference("/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53")


def test_selectively_enforced_services(mock_workload):
    assert len(mock_workload.selectively_enforced_services) == 4
    assert isinstance(mock_workload.selectively_enforced_services[0], Service) \
        and isinstance(mock_workload.selectively_enforced_services[2], ServicePort)


def test_get_by_enforcement_mode(pce):
    workloads = pce.workloads.get(params={'enforcement_mode': 'visibility_only'})
    assert len(workloads) == 1


def test_create_workload(pce, new_workload):
    created_workload = pce.workloads.create(new_workload)
    assert created_workload.href != ''
    workload = pce.workloads.get_by_reference(created_workload.href)
    assert created_workload == workload


def test_update_workload(pce, mock_workload):
    pce.workloads.update(mock_workload.href, {'enforcement_mode': 'selective'})
    updated_workload = pce.workloads.get_by_reference(mock_workload.href)
    assert updated_workload.enforcement_mode == 'selective'



# -*- coding: utf-8 -*-

"""This module is a stub for security principal objects.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import MutableObject, pce_api


@dataclass
@pce_api('security_principals')
class SecurityPrincipal(MutableObject):
    sid: str = None



import pytest

from illumio.policyobjects import IPList, IPRange, FQDN
from illumio.util import DRAFT

from helpers import random_string


@pytest.fixture
def ip_list(pce, session_identifier):
    identifier = random_string()
    ip_list = pce.ip_lists.create(
        IPList(
            name='{}-IPL-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            ip_ranges=[
                IPRange(
                    from_ip='192.168.0.0/16',
                    description='Local network addresses'
                ),
                IPRange(
                    from_ip='10.0.0.1',
                    to_ip='10.0.1.255',
                    description='Private addresses'
                ),
                {'from_ip': '127.0.0.1', 'description': 'Loopback address'},
                {'from_ip': '192.168.0.1', 'exclusion': True, 'description': 'Exclude router address'}
            ],
            fqdns=[
                FQDN(fqdn='*.internal.labs.io', description='Internal lab domains'),
                {'fqdn': 'localhost', 'description': 'Local hostname'}
            ],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield ip_list
    pce.ip_lists.delete(ip_list.href)


def test_get_by_reference(pce, ip_list):
    ipl = pce.ip_lists.get_by_reference(ip_list.href)
    assert ipl == ip_list


def test_get_by_partial_name(pce, session_identifier, ip_list):
    ip_lists = pce.ip_lists.get(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(ip_lists) == 1


def test_get_by_ip(pce, ip_list):
    ip_lists = pce.ip_lists.get(params={'ip_list': '192.168.123.255'}, policy_version=DRAFT)
    assert len(ip_lists) == 2  # should include the default Any IP list


def test_get_by_fqdn(pce, ip_list):
    ip_lists = pce.ip_lists.get(params={'fqdn': 'internal.labs.io'}, policy_version=DRAFT)
    assert len(ip_lists) == 1


def test_get_async(pce, session_identifier, ip_list):
    ip_lists = pce.ip_lists.get_async(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(ip_lists) == 1


def test_update_ip_list(pce, ip_list):
    pce.ip_lists.update(
        ip_list.href,
        {
            'description': 'Integration test update. Remove FQDNs.',
            'fqdns': []
        }
    )
    ipl = pce.ip_lists.get_by_reference(ip_list.href)
    assert ipl.fqdns == []



import json
import os
import re
from typing import List

import pytest

from illumio.policyobjects import IPList, IPRange
from illumio.util import ANY_IP_LIST_NAME, DRAFT, ACTIVE

IP_LISTS = os.path.join(pytest.DATA_DIR, 'ip_lists.json')


@pytest.fixture(scope='module')
def ip_lists() -> List[dict]:
    with open(IP_LISTS, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(scope='module')
def new_ip_list() -> IPList:
    return IPList(
        name="IPL-INTERNAL",
        ip_ranges=[
            IPRange(from_ip='10.0.0.0/8'),
            IPRange(from_ip='172.16.0.0/12'),
            IPRange(from_ip='192.168.0.0', to_ip='192.168.255.255')
        ]
    )


@pytest.fixture(autouse=True)
def ip_lists_mock(pce_object_mock, ip_lists):
    pce_object_mock.add_mock_objects(ip_lists)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    pattern = re.compile('/sec_policy/(draft|active)/ip_lists')
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)


def test_get_default_ip_list(pce):
    default_ip_list = pce.get_default_ip_list()
    assert default_ip_list.name == ANY_IP_LIST_NAME


def test_get_by_reference(pce):
    any_ip_list = pce.ip_lists.get_by_reference('/orgs/1/sec_policy/active/ip_lists/1')
    assert any_ip_list.name == ANY_IP_LIST_NAME


def test_get_by_name(pce):
    ip_lists = pce.ip_lists.get(params={"name": "IPL-"}, policy_version=DRAFT)
    repeated_ip_lists = [ip_list for ip_list in ip_lists if ip_list.name == "IPL-4"]
    assert len(repeated_ip_lists) == 1
    assert '/draft/' in repeated_ip_lists[0].href


def test_get_active_ip_list(pce):
    ip_lists = pce.ip_lists.get(params={"name": "IPL-"}, policy_version=ACTIVE)
    repeated_ip_lists = [ip_list for ip_list in ip_lists if ip_list.name == "IPL-4"]
    assert len(repeated_ip_lists) == 1
    assert '/active/' in repeated_ip_lists[0].href


def test_create_ip_list(pce, new_ip_list):
    ip_list = pce.ip_lists.create(new_ip_list)
    assert ip_list.href != ''
    fetched_ip_list = pce.ip_lists.get_by_reference(ip_list.href)
    assert fetched_ip_list == ip_list


def test_update_ip_list(pce):
    ip_list = pce.ip_lists.get(params={"name": "IPL-4", 'max_results': 1})[0]
    pce.ip_lists.update(ip_list.href, {'fqdns': [{'fqdn': 'test.example.com'}]})
    updated_ip_list = pce.ip_lists.get_by_reference(ip_list.href)
    assert len(updated_ip_list.fqdns) > 0
    assert updated_ip_list.fqdns[0].fqdn == 'test.example.com'



# -*- coding: utf-8 -*-

"""This module provides classes for traffic analysis using Explorer.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import re
import socket
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, List, Union, Optional

from illumio import IllumioException
from illumio.infrastructure import Network
from illumio.policyobjects import (
    IPList,
    Label,
    ServicePort,
    VirtualServer,
    VirtualService
)
from illumio.util import (
    JsonObject,
    Reference,
    Transmission,
    PolicyDecision,
    FlowDirection,
    TrafficState,
    FQDN_REGEX
)
from illumio.workloads import Workload

AND = 'and'
OR = 'or'


@dataclass
class TrafficQueryFilter(JsonObject):
    label: Reference = None
    workload: Reference = None
    ip_list: Reference = None
    ip_address: str = None
    fqdn: str = None
    transmission: str = None

    def _validate(self):
        if self.transmission and not self.transmission in Transmission:
            raise IllumioException("Invalid transmission: {}".format(self.transmission))
        super()._validate()


@dataclass
class TrafficQueryFilterBlock(JsonObject):
    # the include parameter is specified as a list of lists
    # of object references or key-value pairs
    include: List[List[TrafficQueryFilter]] = field(default_factory=list)
    exclude: List[TrafficQueryFilter] = field(default_factory=list)

    def __post_init__(self):
        self.include = _parse_traffic_filters(self.include, include=True)
        self.exclude = _parse_traffic_filters(self.exclude)
        super().__post_init__()


def _parse_traffic_filters(refs: List[Any], include=False) -> List[object]:
    traffic_objects = []
    for ref in refs:
        if type(ref) is not str:
            traffic_objects.append(ref)
            continue
        elif re.match(FQDN_REGEX, ref):
            if include:
                raise IllumioException("Cannot specify consumer FQDN filter")
            o = {'fqdn': ref}
        elif 'label' in ref:
            o = {'label': {'href': ref}}
        elif 'workload' in ref:
            o = {'workload': {'href': ref}}
        elif 'ip_list' in ref:
            o = {'ip_list': {'href': ref}}
        elif ref in Transmission:
            if include:
                raise IllumioException("Cannot specify consumer transmission filter")
            o = {'transmission': ref}
        else:
            try:
                socket.inet_aton(ref)  # check if the reference is an IP address
                o = {'ip_address': ref}
            except socket.error:
                raise IllumioException('Invalid traffic filter type: {}').format(ref)
        traffic_objects.append([o] if include else o)
    return traffic_objects


@dataclass
class TrafficQueryServiceBlock(JsonObject):
    include: List[ServicePort] = field(default_factory=list)
    exclude: List[ServicePort] = field(default_factory=list)

    def __post_init__(self):
        self.include = _parse_service_ports(self.include)
        self.exclude = _parse_service_ports(self.exclude)
        super().__post_init__()


def _parse_service_ports(service_ports: List[Union[ServicePort, dict]]):
    parsed_service_ports = []
    for service_port in service_ports:
        if type(service_port) is ServicePort:
            parsed_service_ports.append(service_port)
        elif type(service_port) is dict:
            parsed_service_ports.append(ServicePort.from_json(service_port))
        else:
            raise IllumioException("Invalid service port type: {}".format(type(service_port)))
    return parsed_service_ports


@dataclass
class TrafficQuery(JsonObject):
    start_date: Union[str, int, float]
    end_date: Union[str, int, float]
    sources: TrafficQueryFilterBlock = field(default_factory=TrafficQueryFilterBlock)
    destinations: TrafficQueryFilterBlock = field(default_factory=TrafficQueryFilterBlock)
    services: TrafficQueryServiceBlock = field(default_factory=TrafficQueryServiceBlock)
    policy_decisions: List[str] = field(default_factory=list)
    exclude_workloads_from_ip_list_query: bool = True
    sources_destinations_query_op: str = AND
    max_results: int = 100000
    query_name: str = None  # required for async traffic queries

    @staticmethod
    def build(start_date: Optional[Union[str, int, float]] = None,
            end_date: Optional[Union[str, int, float]] = None,
            include_sources=[], exclude_sources=[],
            include_destinations=[], exclude_destinations=[],
            include_services=[], exclude_services=[], policy_decisions=[],
            exclude_workloads_from_ip_list_query=True, max_results=100000,
            query_name=None) -> 'TrafficQuery':
        return TrafficQuery(
            start_date=start_date, end_date=end_date,
            sources=TrafficQueryFilterBlock(
                include=include_sources,
                exclude=exclude_sources
            ),
            destinations=TrafficQueryFilterBlock(
                include=include_destinations,
                exclude=exclude_destinations
            ),
            services=TrafficQueryServiceBlock(
                include=include_services,
                exclude=exclude_services
            ),
            policy_decisions=policy_decisions,
            exclude_workloads_from_ip_list_query=exclude_workloads_from_ip_list_query,
            max_results=max_results, query_name=query_name
        )

    def __post_init__(self):
        if type(self.start_date) is int or type(self.start_date) is float:
            self.start_date = self._convert_timestamp_to_date_string(self.start_date)
        if type(self.end_date) is int or type(self.end_date) is float:
            self.end_date = self._convert_timestamp_to_date_string(self.end_date)
        super().__post_init__()

    def _convert_timestamp_to_date_string(self, timestamp: Union[int, float]) -> str:
        try:
            # the Unix timestamp could be in seconds or milliseconds,
            # so check the number of digits; 12 digits in s is year 5138
            # and we could theoretically be looking at dates before 2001 (10 digits)
            if len(str(int(timestamp))) >= 12:
                timestamp = timestamp / 1000
            dt = datetime.utcfromtimestamp(timestamp)
            return dt.strftime('%Y-%m-%dT%H:%M:%SZ')
        except Exception:
            raise IllumioException("Invalid start or end time provided for traffic analysis")

    def _validate(self):
        for policy_decision in self.policy_decisions:
            if not policy_decision in PolicyDecision:
                raise IllumioException("Invalid policy_decision: {}".format(policy_decision))
        if self.sources_destinations_query_op.lower() not in {AND, OR}:
            raise IllumioException("sources_destinations_query_op must be one of 'and' or 'or', was {}".format(self.sources_destinations_query_op))
        super()._validate()


@dataclass
class TrafficNode(JsonObject):
    ip: str = None
    label: Label = None
    workload: Workload = None
    ip_lists: List[IPList] = None
    virtual_server: VirtualServer = None
    virtual_service: VirtualService = None


@dataclass
class TimestampRange(JsonObject):
    first_detected: str
    last_detected: str


@dataclass
class TrafficFlow(JsonObject):
    src: TrafficNode
    dst: TrafficNode
    service: ServicePort = None
    num_connections: int = None
    state: str = None
    timestamp_range: TimestampRange = None
    dst_bi: int = None
    dst_bo: int = None
    policy_decision: str = None
    flow_direction: str = None
    transmission: str = None
    icmp_type: int = None
    icmp_code: int = None
    network: Network = None

    def _validate(self):
        if self.flow_direction and not self.flow_direction in FlowDirection:
            raise IllumioException("Invalid flow_direction: {}".format(self.flow_direction))
        if self.policy_decision and not self.policy_decision in PolicyDecision:
            raise IllumioException("Invalid policy_decision: {}".format(self.policy_decision))
        if self.state and not self.state in TrafficState:
            raise IllumioException("Invalid state: {}".format(self.state))
        if self.transmission and not self.transmission in Transmission:
            raise IllumioException("Invalid transmission: {}".format(self.transmission))
        super()._validate()



# -*- coding: utf-8 -*-

"""This module provides classes related to policy rules.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List, Union

from illumio.util import JsonObject, Reference, MutableObject, pce_api
from illumio.policyobjects import Service, ServicePort

from .actor import Actor


@dataclass
class BaseRule(JsonObject):
    ingress_services: List[Union[Service, ServicePort]] = None
    providers: List[Actor] = None
    consumers: List[Actor] = None

    @classmethod
    def build(cls, providers: List[Union[str, Reference, dict]], consumers: List[Union[str, Reference, dict]],
            ingress_services: List[Union[JsonObject, dict, str]], **kwargs) -> 'BaseRule':
        services = []
        for service in ingress_services:
            if isinstance(service, JsonObject):
                services.append(service)
            elif type(service) is str:
                services.append(Service(href=service))
            else:
                service_type = Service if 'href' in service else ServicePort
                services.append(service_type.from_json(service))
        return cls(
            providers=[Actor.from_reference(provider) for provider in providers],
            consumers=[Actor.from_reference(consumer) for consumer in consumers],
            ingress_services=services,
            **kwargs
        )

    def _decode_complex_types(self):
        decoded_ingress_services = []
        if self.ingress_services:
            for service in self.ingress_services:
                service_type = Service if 'href' in service else ServicePort
                decoded_ingress_services.append(service_type.from_json(service))
        self.ingress_services = decoded_ingress_services
        super()._decode_complex_types()


@dataclass
class LabelResolutionBlock(JsonObject):
    providers: List[str] = None
    consumers: List[str] = None


@dataclass
@pce_api('rules', endpoint='/sec_rules')
class Rule(BaseRule, MutableObject):
    """Represents a security rule in the PCE.

    Each security rule defines one or more services on which traffic will be
    allowed from the defined providers to the defined consumers.

    Providers and consumers can be defined using static (workload HREF) or
    dynamic (label, IP list) references. By default, providers and consumers
    are resolved as workloads.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/create-security-policy/rules.htm

    Usage:
        >>> from illumio import PolicyComputeEngine, Rule, RuleSet, LabelSet
        >>> any_ip_list = pce.get_default_ip_list()
        >>> role_label = pce.labels.create({'key': 'role', 'value': 'Web'})
        >>> app_label = pce.labels.create({'key': 'app', 'value': 'App'})
        >>> env_label = pce.labels.create({'key': 'env', 'value': 'Production'})
        >>> loc_label = pce.labels.create({'key': 'loc', 'value': 'AWS'})
        >>> ruleset = RuleSet(
        ...     name='RS-LAB-ALLOWLIST',
        ...     scopes=[
        ...         LabelSet(
        ...             labels=[app_label, env_label, loc_label]
        ...         )
        ...     ]
        ... )
        >>> ruleset = pce.rule_sets.create(ruleset)
        >>> rule = Rule.build(
        ...     providers=[role_label],
        ...     consumers=[any_ip_list],
        ...     ingress_services=[
        ...         {'port': 80, 'proto': 'tcp'},
        ...         {'port': 443, 'proto': 'tcp'}
        ...     ],
        ...     resolve_providers_as=['workloads'],
        ...     resolve_consumers_as=['workloads'],
        ...     unscoped_consumers=True  # creates an extra-scope rule
        ... )
        >>> rule = pce.rules.create(rule, parent=ruleset)
        >>> rule
        Rule(
            href='/orgs/1/sec_policy/rule_sets/19/rules/sec_rules/34',
            enabled=True,
            providers=[
                Actor(
                    label=Reference(
                        href='/orgs/1/labels/21'
                    ),
                    ...
                )
            ],
            consumers=[
                Actor(
                    ip_list=Reference(
                        href='/orgs/1/sec_policy/draft/ip_lists/1'
                    ),
                    ...
                )
            ],
            ingress_services=[
                ServicePort(port=80, proto=6, ...),
                ServicePort(port=443, proto=6, ...)
            ],
            resolve_labels_as=LabelResolutionBlock(
                providers=['workloads'],
                consumers=['workloads']
            ),
            unscoped_consumers=True,
            ...
        )
    """
    enabled: bool = None
    resolve_labels_as: LabelResolutionBlock = None
    sec_connect: bool = None
    stateless: bool = None
    machine_auth: bool = None
    consuming_security_principals: List[Reference] = None
    unscoped_consumers: bool = None
    network_type: str = None

    @classmethod
    def build(cls, providers: List[Union[str, Reference, dict]], consumers: List[Union[str, Reference, dict]],
            ingress_services: List[Union[JsonObject, dict, str]],
            resolve_providers_as: List[str], resolve_consumers_as: List[str], enabled=True, **kwargs) -> 'Rule':
        resolve_labels_as = LabelResolutionBlock(providers=resolve_providers_as, consumers=resolve_consumers_as)
        return super().build(providers, consumers, ingress_services, resolve_labels_as=resolve_labels_as, enabled=enabled, **kwargs)



# -*- coding: utf-8 -*-

"""This module provides classes related to IP list policy objects.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio.util import JsonObject, MutableObject, pce_api


@dataclass
class IPRange(JsonObject):
    description: str = None
    from_ip: str = None
    to_ip: str = None
    exclusion: bool = None


@dataclass
class FQDN(JsonObject):
    fqdn: str = None
    description: str = None


@dataclass
@pce_api('ip_lists', is_sec_policy=True)
class IPList(MutableObject):
    """Represents an IP list in the PCE.

    IP lists are list of IP addresses, subnets, CIDR blocks, and/or FQDNs.
    They can be used in conjunction with other security policy objects to
    allow or deny traffic from these defined ranges.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/ip-lists.htm

    Usage:
        >>> from illumio import PolicyComputeEngine, IPList, IPRange
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> pce.set_credentials('api_key_username', 'api_key_secret')
        >>> ip_list = IPList(
        ...     name='IPL-INTERNAL',
        ...     ip_ranges=IPRange(
        ...         from_ip='192.168.0.0/16'
        ...     )
        ... )
        >>> ip_list = pce.ip_lists.create(ip_list)
        >>> ip_list
        IPList(
            href='/orgs/1/sec_policy/draft/ip_lists/22',
            name='IPL-INTERNAL',
            ip_ranges=IPRange(
                from_ip='192.168.0.0/16'
            ),
            ...
        )
    """
    ip_ranges: List[IPRange] = None
    fqdns: List[FQDN] = None



import os

import pytest

TEST_DIR = os.path.abspath(os.path.dirname(__file__))


def pytest_addoption(parser):
    parser.addoption("--integration", action="store_true", default=False, help="run integration tests")


def pytest_configure(config):
    pytest.DATA_DIR = os.path.join(TEST_DIR, 'data')
    config.addinivalue_line("markers", "integration: mark integration tests")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--integration"):
        return
    integration_skip_marker = pytest.mark.skip(reason="use --integration marker to run")
    for item in items:
        filename = str(item.fspath).split(os.path.sep)[-1]
        if "integration" in item.keywords or filename.startswith('test_intg'):
            item.add_marker(integration_skip_marker)



# -*- coding: utf-8 -*-

"""This module provides classes for security policy provisioning.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import re
from dataclasses import dataclass
from typing import List

from .exceptions import IllumioException
from .util import (
    JsonObject,
    Reference,
    MutableObject,
    ImmutableObject,
    HREF_REGEX
)
from .policyobjects import LabelSet


@dataclass
class FirewallSetting(MutableObject):
    allow_dhcp_client: bool = None
    log_dropped_multicast: bool = None
    log_dropped_broadcast: bool = None
    allow_traceroute: bool = None
    allow_ipv6: bool = None
    ipv6_mode: str = None
    network_detection_mode: str = None
    ike_authentication_type: str = None
    static_policy_scopes: List[LabelSet] = None
    firewall_coexistence: List[LabelSet] = None
    containers_inherit_host_policy_scopes: List[LabelSet] = None
    blocked_connection_reject_scopes: List[LabelSet] = None
    loopback_interfaces_in_policy_scopes: List[LabelSet] = None


@dataclass
class PolicyChangeset(JsonObject):
    label_groups: List[Reference] = None
    services: List[Reference] = None
    rule_sets: List[Reference] = None
    ip_lists: List[Reference] = None
    virtual_services: List[Reference] = None
    firewall_settings: List[Reference] = None
    enforcement_boundaries: List[Reference] = None
    secure_connect_gateways: List[Reference] = None
    virtual_servers: List[Reference] = None

    @staticmethod
    def build(hrefs: List[str]):
        changeset = PolicyChangeset()
        for href in hrefs:
            match = re.match(HREF_REGEX, href)
            if match:
                object_type = match.group('type')
                arr = getattr(changeset, object_type) or []
                arr.append(Reference(href=href))
                setattr(changeset, object_type, arr)
            else:
                raise IllumioException('Invalid HREF in policy provision changeset: {}'.format(href))
        return changeset


@dataclass
class PolicyObjectCounts(JsonObject):
    label_groups: int = None
    services: int = None
    rule_sets: int = None
    ip_lists: int = None
    virtual_services: int = None
    firewall_settings: int = None
    enforcement_boundaries: int = None
    secure_connect_gateways: int = None
    virtual_servers: int = None


@dataclass
class PolicyVersion(ImmutableObject):
    commit_message: str = None
    version: int = None
    workloads_affected: int = None
    object_counts: PolicyObjectCounts = None


__all__ = [
    'FirewallSetting',
    'PolicyVersion',
    'PolicyObjectCounts',
    'PolicyChangeset'
]



import pytest

from illumio.util import EnforcementMode, VisibilityLevel
from illumio.workloads import PairingProfile

from helpers import random_string


@pytest.fixture
def pairing_profile(pce, session_identifier, env_label):
    identifier = random_string()
    pairing_profile = pce.pairing_profiles.create(
        PairingProfile(
            name='{}-PP-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            enabled=True,
            # agent_software_release: str = None
            enforcement_mode=EnforcementMode.VISIBILITY_ONLY.value,
            enforcement_mode_lock=True,
            visibility_level=VisibilityLevel.FLOW_SUMMARY.value,
            visibility_level_lock=True,
            allowed_uses_per_key=1,
            key_lifespan='unlimited',
            labels=[env_label],
            role_label_lock=False,
            app_label_lock=False,
            env_label_lock=True,
            loc_label_lock=False,
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield pairing_profile
    pce.pairing_profiles.delete(pairing_profile.href)


def test_get_by_reference(pce, pairing_profile):
    profile = pce.pairing_profiles.get_by_reference(pairing_profile.href)
    assert profile == pairing_profile


def test_get_by_partial_name(pce, session_identifier, pairing_profile):
    pairing_profiles = pce.pairing_profiles.get(params={'name': session_identifier})
    assert len(pairing_profiles) == 1


def test_update_pairing_profile(pce, pairing_profile):
    pce.pairing_profiles.update(pairing_profile.href, {
        'enforcement_mode': 'selective'
    })
    profile = pce.pairing_profiles.get_by_reference(pairing_profile.href)
    assert pairing_profile.href == profile.href
    assert profile.enforcement_mode == 'selective'



# -*- coding: utf-8 -*-

"""This module provides common exception classes used throughout the application.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
class IllumioException(Exception):
    """Superclass for Illumio library exceptions"""


class IllumioApiException(IllumioException):
    """Superclass for exceptions generated by the Illumio API code"""



from illumio.policyobjects import ServicePort
from illumio.util import convert_protocol, DRAFT


def test_get_by_reference(pce, web_service):
    service = pce.services.get_by_reference(web_service.href)
    assert service == web_service


def test_get_by_partial_name(pce, session_identifier, web_service, well_known_service):
    services = pce.services.get(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(services) == 2


def test_get_by_port(pce, session_identifier, web_service, well_known_service, rdp_service):
    services = pce.services.get(params={'port': 443, 'name': session_identifier}, policy_version=DRAFT)
    assert len(services) == 2


def test_get_by_proto(pce, session_identifier, web_service, well_known_service, rdp_service):
    services = pce.services.get(params={'proto': convert_protocol('udp'), 'name': session_identifier}, policy_version=DRAFT)
    assert len(services) == 1


def test_get_async(pce, session_identifier, web_service, well_known_service, rdp_service):
    services = pce.services.get_async(params={'name': session_identifier})
    assert len(services) == 3


def test_update_service(pce, well_known_service):
    updated_service_ports = well_known_service.service_ports
    updated_service_ports.append(ServicePort.from_json({'port': 1, 'to_port': 1023, 'proto': 'udp'}))
    pce.services.update(well_known_service.href, {'service_ports': updated_service_ports})
    service = pce.services.get_by_reference(well_known_service.href)
    assert service.service_ports == updated_service_ports



import json
import os
import re
from typing import List

import pytest

from illumio.workloads import PairingProfile

MOCK_PAIRING_PROFILES = os.path.join(pytest.DATA_DIR, 'pairing_profiles.json')
MOCK_PAIRING_KEY = os.path.join(pytest.DATA_DIR, 'pairing_key.json')


@pytest.fixture(scope='module')
def pairing_profiles() -> List[dict]:
    with open(MOCK_PAIRING_PROFILES, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(scope='module')
def mock_pairing_key() -> dict:
    with open(MOCK_PAIRING_KEY, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(autouse=True)
def pairing_profiles_mock(pce_object_mock, pairing_profiles):
    pce_object_mock.add_mock_objects(pairing_profiles)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback, mock_pairing_key):
    profile_pattern = re.compile('/pairing_profiles')
    requests_mock.register_uri('GET', profile_pattern, json=get_callback)
    requests_mock.register_uri('POST', profile_pattern, json=post_callback)
    requests_mock.register_uri('PUT', profile_pattern, json=put_callback)
    requests_mock.register_uri('DELETE', profile_pattern, json=delete_callback)

    key_pattern = re.compile('/pairing_key')
    requests_mock.register_uri('POST', key_pattern, json=mock_pairing_key)


def test_get_profiles_by_name(pce):
    pairing_profiles = pce.pairing_profiles.get(params={'name': 'PP-'})
    assert len(pairing_profiles) == 2


def test_get_by_labels(pce):
    params = {'labels': '[["/orgs/1/labels/136"]]'}
    pairing_profiles = pce.pairing_profiles.get(params=params)
    assert len(pairing_profiles) == 2


def test_get_by_multiple_labels(pce):
    params = {'labels': '[["/orgs/1/labels/224","/orgs/1/labels/136"]]'}
    pairing_profiles = pce.pairing_profiles.get(params=params)
    assert len(pairing_profiles) == 1


def test_get_by_distinct_labels(pce):
    params = {'labels': '[["/orgs/1/labels/224"],["/orgs/1/labels/136"]]'}
    pairing_profiles = pce.pairing_profiles.get(params=params)
    assert len(pairing_profiles) == 3


def test_pairing_profile_create(pce):
    pairing_profile = PairingProfile(name='PP-TEST', enabled=True)
    pairing_profile = pce.pairing_profiles.create(pairing_profile)
    pairing_profiles = pce.pairing_profiles.get()
    assert len(pairing_profiles) == 4
    assert pairing_profile.href in [p.href for p in pairing_profiles]


def test_pairing_profile_update(pce):
    test_description = 'Updated test profile'
    pairing_profiles = pce.pairing_profiles.get()
    pairing_profile = pairing_profiles[0]
    pairing_profile.description = test_description
    pce.pairing_profiles.update(pairing_profile.href, pairing_profile)
    pairing_profile = pce.pairing_profiles.get_by_reference(pairing_profile.href)
    assert pairing_profile.description == test_description


def test_pairing_key_generation(pce, mock_pairing_key):
    pairing_profiles = pce.pairing_profiles.get()
    pairing_key = pce.generate_pairing_key(pairing_profiles[0].href)
    assert pairing_key == mock_pairing_key.get('activation_code')



# -*- coding: utf-8 -*-

"""This module provides classes related to RBAC users in the PCE.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import IllumioObject


@dataclass
class User(IllumioObject):
    username: str = None
    last_login_on: str = None
    last_login_ip_address: str = None
    login_count: int = None
    full_name: str = None
    time_zone: str = None
    locked: bool = None
    effective_groups: list = None
    local_profile: dict = None
    type: str = None
    presence_status: str = None



import json
import os
import re
from typing import List

import pytest

from illumio.rules import RuleSet
from illumio.util import DRAFT, ACTIVE

RULESETS = os.path.join(pytest.DATA_DIR, 'rule_sets.json')


@pytest.fixture(scope='module')
def rule_sets() -> List[dict]:
    with open(RULESETS, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(scope='module')
def new_rule_set() -> RuleSet:
    return RuleSet(
        name="RS-TEST"
    )


@pytest.fixture(autouse=True)
def rule_sets_mock(pce_object_mock, rule_sets):
    pce_object_mock.add_mock_objects(rule_sets)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    pattern = re.compile('/sec_policy/(draft|active)/rule_sets')
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)


@pytest.fixture()
def mock_rule_set(pce) -> RuleSet:
    yield pce.rule_sets.get_by_reference("/orgs/1/sec_policy/active/rule_sets/1")


def test_encoded_scopes(mock_rule_set):
    json_rule_set = mock_rule_set.to_json()
    assert json_rule_set['scopes'] == [[]]


def test_get_by_name(pce):
    rule_sets = pce.rule_sets.get(params={'name': 'RS-'}, policy_version=DRAFT)
    assert len(rule_sets) == 2


def test_get_active_rule_sets(pce, mock_rule_set):
    rule_set = pce.rule_sets.get(params={'name': 'RS-RINGFENCE', 'max_results': 1}, policy_version=ACTIVE)[0]
    assert rule_set == mock_rule_set


def test_create_rule_set(pce, new_rule_set):
    created_rule_set = pce.rule_sets.create(new_rule_set)
    assert created_rule_set.href != ''
    rule_set = pce.rule_sets.get_by_reference(created_rule_set.href)
    assert created_rule_set == rule_set


def test_update_rule_set(pce, mock_rule_set):
    pce.rule_sets.update(mock_rule_set.href, {'enabled': False})
    updated_rule_set = pce.rule_sets.get_by_reference(mock_rule_set.href)
    assert updated_rule_set.enabled is False



import pytest

from illumio.util import (
    convert_draft_href_to_active,
    convert_protocol,
    ACTIVE
)

from helpers import random_string


@pytest.mark.parametrize(
    "api_name,sec_policy_object", [
        (
            'ip_lists',
            {
                'description': 'Test IP List provisioning',
                'ip_ranges': [{'from_ip': '10.0.0.0/8'}],
            }
        ),
        (
            'label_groups',
            {
                'key': 'env',
                'description': 'Created by illumio python library integration tests',
                'labels': []
            }
        ),
        (
            'services',
            {
                'description': 'Test service provisioning',
                'service_ports': [{'port': 3306, 'proto': convert_protocol('tcp')}]
            }
        ),
        (
            'rule_sets',
            {
                'description': 'Test rule set provisioning',
                'enabled': True,
                'scopes': []
            }
        )
    ]
)
def test_provision_object(pce, api_name, sec_policy_object, session_identifier, request):
    identifier = random_string()
    api = getattr(pce, api_name)
    sec_policy_object.update({
        'name': '{}-{}'.format(session_identifier, identifier),
        'external_data_set': session_identifier,
        'external_data_reference': identifier
    })
    policy_object = api.create(sec_policy_object)
    pce.provision_policy_changes(
        change_description='Test {} object provisioning'.format(api_name),
        hrefs=[policy_object.href]
    )

    def _teardown():
        api.delete(policy_object.href)
        pce.provision_policy_changes(
            change_description='Remove provisioned {} object'.format(api_name),
            hrefs=[policy_object.href]
        )

    request.addfinalizer(_teardown)

    policy_objects = api.get(params={'name': session_identifier}, policy_version=ACTIVE)
    assert len(policy_objects) == 1 and policy_objects[0].href == convert_draft_href_to_active(policy_object.href)



import json
import os
import re
from datetime import datetime, timezone
from typing import List

import pytest
from requests_mock import ANY

from illumio import IllumioException
from illumio.explorer import TrafficQuery, TrafficQueryFilterBlock, TrafficFlow

MOCK_TRAFFIC_QUERY = os.path.join(pytest.DATA_DIR, 'traffic_query.json')
MOCK_TRAFFIC_FLOWS = os.path.join(pytest.DATA_DIR, 'traffic_query_response.json')


@pytest.fixture(scope='module')
def traffic_query() -> TrafficQuery:
    with open(MOCK_TRAFFIC_QUERY, 'r') as f:
        yield TrafficQuery.from_json(f.read())


@pytest.fixture(scope='module')
def traffic_flows() -> List[TrafficFlow]:
    with open(MOCK_TRAFFIC_FLOWS, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(autouse=True)
def traffic_flows_mock(pce_object_mock, traffic_flows):
    pce_object_mock.add_mock_objects(traffic_flows)


@pytest.fixture
def traffic_query_callback(pce_object_mock):
    def _callback_fn(request, context):
        return pce_object_mock.get_mock_objects(request.path_url)
    return _callback_fn


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, traffic_query_callback):
    pattern = re.compile('/traffic_flows/traffic_analysis_queries')
    requests_mock.register_uri('POST', pattern, json=traffic_query_callback)


def test_query_structure(traffic_query):
    assert type(traffic_query.sources) is TrafficQueryFilterBlock


def test_traffic_query(pce, traffic_query):
    traffic_flows = pce.get_traffic_flows(traffic_query)
    assert len(traffic_flows) > 0
    assert traffic_flows[0].src is not None


def test_invalid_policy_decision():
    with pytest.raises(IllumioException):
        TrafficQuery(start_date='2021-11-05', end_date='2021-11-12', policy_decisions=["invalid_policy_decision"])


def test_timestamp_conversion():
    start_date = datetime.strptime('2021-11-05', '%Y-%m-%d').replace(tzinfo=timezone.utc)
    start_time_seconds = start_date.timestamp()
    end_date = datetime.strptime('2021-11-12', '%Y-%m-%d').replace(tzinfo=timezone.utc)
    end_time_milliseconds = end_date.timestamp() * 1000
    query = TrafficQuery(start_date=start_time_seconds, end_date=end_time_milliseconds, policy_decisions=["unknown"])
    assert query.start_date == '2021-11-05T00:00:00Z'
    assert query.end_date == '2021-11-12T00:00:00Z'



import json

import pytest

from illumio import PolicyComputeEngine
from mocks import PCEObjectMock


@pytest.fixture(scope='session')
def pce():
    return PolicyComputeEngine('test.pce.com')


@pytest.fixture
def pce_object_mock():
    yield PCEObjectMock()


@pytest.fixture
def get_callback(pce_object_mock):
    def _callback_fn(request, context):
        context.headers['X-Total-Count'] = str(len(pce_object_mock.mock_objects))
        return pce_object_mock.get_mock_objects(request.path_url)
    return _callback_fn


@pytest.fixture
def post_callback(pce_object_mock):
    def _callback_fn(request, context):
        json_body = json.loads(request.body.decode('utf-8'))
        return pce_object_mock.create_mock_object(request.path_url, json_body)
    return _callback_fn


@pytest.fixture
def put_callback(pce_object_mock):
    def _callback_fn(request, context):
        json_body = json.loads(request.body.decode('utf-8'))
        pce_object_mock.update_mock_object(request.path_url, json_body)
    return _callback_fn


@pytest.fixture
def delete_callback(pce_object_mock):
    def _callback_fn(request, context):
        pce_object_mock.delete_mock_object(request.path_url)
    return _callback_fn



# -*- coding: utf-8 -*-

"""This module is a stub for enforcement boundary objects.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import pce_api

from .rule import BaseRule


@dataclass
@pce_api('enforcement_boundaries', is_sec_policy=True)
class EnforcementBoundary(BaseRule):
    """Represents an enforcement boundary in the PCE.

    Enforcement boundaries establish deny rules for workloads within their scope
    communicating on its defined services.

    Workloads in selective enforcement mode that fall within an enforcement
    boundary will have policy rules apply to them as if they were in full
    enforcement.

    Rules allowing traffic that would otherwise be denied by an enforcement
    boundary will override the boundary's deny rule.

    Usage:
        >>> from illumio import PolicyComputeEngine, EnforcementBoundary, AMS
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> pce.set_credentials('api_key_username', 'api_key_secret')
        >>> any_ip_list = pce.get_default_ip_list()
        >>> enforcement_boundary = EnforcementBoundary.build(
        ...     name='EB-BLOCK-RDP',
        ...     providers=[AMS],  # the special 'ams' literal denotes all workloads
        ...     consumers=[any_ip_list.href],
        ...     ingress_services=[
        ...         {'port': 3389, 'proto': 'tcp'},
        ...         {'port': 3389, 'proto': 'udp'},
        ...     ]
        ... )
        >>> enforcement_boundary = pce.enforcement_boundaries.create(enforcement_boundary)
        >>> enforcement_boundary
        EnforcementBoundary(
            href='/orgs/1/sec_policy/draft/enforcement_boundary/8',
            name='EB-BLOCK-RDP',
            providers=[
                Actor(
                    actors='ams',
                    ...
                )
            ],
            consumers=[
                Actor(
                    ip_list=Reference(
                        href='/orgs/1/sec_policy/active/ip_lists/1'
                    ),
                    ...
                )
            ],
            ingress_services=[
                ServicePort(port=3389, proto=6),
                ServicePort(port=3389, proto=17)
            ],
            ...
        )
    """
    name: str = None



# -*- coding: utf-8 -*-

"""This module is a stub for network objects.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import MutableObject


@dataclass
class Network(MutableObject):
    pass



from helpers import random_string


def test_get_by_reference(pce, role_label):
    label = pce.labels.get_by_reference(role_label.href)
    assert label == role_label


def test_get_by_partial_name(pce, session_identifier, role_label, app_label, env_label, loc_label):
    labels = pce.labels.get(params={'value': session_identifier})
    assert len(labels) == 4


def test_get_by_key(pce, session_identifier, role_label, app_label):
    labels = pce.labels.get(params={'key': 'role', 'value': session_identifier})
    assert len(labels) == 1


def test_get_async(pce, session_identifier, role_label):
    labels = pce.labels.get_async(params={'value': session_identifier})
    assert len(labels) == 1


def test_create_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create({
        'key': 'role',
        'value': '{}-R-{}'.format(session_identifier, identifier),
        'external_data_set': session_identifier,
        'external_data_reference': identifier
    })
    assert label.href
    pce.labels.delete(label.href)


def test_update_label(pce, session_identifier, env_label):
    updated_value = '{}-E-UPDATED'.format(session_identifier)
    pce.labels.update(
        env_label.href,
        {
            'value': updated_value
        }
    )
    label = pce.labels.get_by_reference(env_label.href)
    assert label.value == updated_value



import json
import os
import re
from collections import namedtuple

import pytest

from illumio import PolicyComputeEngine
from illumio.rules import Rule
from illumio.util import PCE_APIS, DRAFT, ACTIVE


def test_custom_protocol():
    pce = PolicyComputeEngine('http://my.pce.com')
    assert pce.base_url == 'http://my.pce.com:443/api/v2'


def test_invalid_protocol():
    pce = PolicyComputeEngine('ftp://my.pce.com')
    assert pce.base_url == 'https://my.pce.com:443/api/v2'


def test_protocol_parsing():
    pce = PolicyComputeEngine('httpslab.pce.com')
    assert pce.base_url == 'https://httpslab.pce.com:443/api/v2'


def test_path_parsing():
    pce = PolicyComputeEngine('my.pce.com/api/v2')
    assert pce.base_url == 'https://my.pce.com:443/api/v2'


def test_int_org_id():
    pce = PolicyComputeEngine('my.pce.com', org_id=1)
    assert pce.base_url == 'https://my.pce.com:443/api/v2'


_API = namedtuple('API', [
    'name',
    'endpoint',
    'object_class',
    'is_sec_policy',
    'is_global'
])


@pytest.mark.parametrize(
    "api,policy_version,parent,expected", [
        (
            _API(name='rules', endpoint='/sec_rules', object_class=Rule, is_sec_policy=True, is_global=False),
            DRAFT,
            '/orgs/1/sec_policy/active/rule_sets/1',
            '/orgs/1/sec_policy/draft/rule_sets/1/sec_rules'
        ),
        (
            _API(name='rules', endpoint='/sec_rules', object_class=Rule, is_sec_policy=True, is_global=False),
            ACTIVE,
            '/orgs/1/sec_policy/active/rule_sets/1',
            '/orgs/1/sec_policy/draft/rule_sets/1/sec_rules'
        )
    ]
)
def test_endpoint_building(api, policy_version, parent, expected, pce):
    endpoint = PolicyComputeEngine._PCEObjectAPI(pce, api)._build_endpoint(policy_version, parent)
    assert endpoint == expected


@pytest.mark.parametrize(
    "endpoint,include_org,expected", [
        ('/health', False, 'https://test.pce.com:443/api/v2/health'),
        ('labels', True, 'https://test.pce.com:443/api/v2/orgs/1/labels'),
        ('/labels', True, 'https://test.pce.com:443/api/v2/orgs/1/labels'),
        ('//labels', True, 'https://test.pce.com:443/api/v2/orgs/1/labels'),
        ('/sec_policy/active//ip_lists', True, 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/active/ip_lists'),
        ('/sec_policy/active//ip_lists', True, 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/active/ip_lists'),
        ('/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53', True, 'https://test.pce.com:443/api/v2/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53'),
        ('/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53', False, 'https://test.pce.com:443/api/v2/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53'),
        ('/orgs/1/sec_policy/rule_sets/1/sec_rules/1', True, 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/rule_sets/1/sec_rules/1')
    ]
)
def test_url_building(endpoint, include_org, expected, pce):
    assert pce._build_url(endpoint, include_org) == expected


def load_data_file(type_):
    filename = os.path.join(pytest.DATA_DIR, '{}.json'.format(type_))
    with open(filename, 'r') as f:
        return json.loads(f.read())


@pytest.mark.parametrize(
    "api_name,endpoint,ObjectClass,is_sec_policy",
    [
        (name, api.endpoint, api.object_class, api.is_sec_policy)
        for name, api in PCE_APIS.items()
    ]
)
def test_pce_apis(api_name, endpoint, ObjectClass, is_sec_policy, pce,
        requests_mock, get_callback, post_callback, put_callback, delete_callback):
    api = getattr(pce, api_name)

    pattern = re.compile(endpoint)
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)

    obj = api.create(ObjectClass(name='test object'))
    assert obj.href

    policy_version = DRAFT if is_sec_policy else None
    assert api.get_by_reference(obj.href) == obj

    objs = api.get(params={'name': 'test object'}, policy_version=policy_version)
    assert objs[0] == obj

    api.update(obj.href, {'description': 'updated description'})
    obj = api.get_by_reference(obj.href)
    assert obj.description == 'updated description'

    api.delete(obj.href)
    objs = api.get(policy_version=policy_version)
    assert len(objs) == 0



# -*- coding: utf-8 -*-

"""This module provides classes related to policy rule sets.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio.util import MutableObject, pce_api
from illumio.policyobjects import LabelSet

from .rule import Rule
from .iptablesrule import IPTablesRule


@dataclass
@pce_api('rule_sets', is_sec_policy=True)
class RuleSet(MutableObject):
    """Provides scope boundaries for security policy rules.

    Usage:
        >>> from illumio import PolicyComputeEngine, RuleSet, LabelSet
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> pce.set_credentials('api_key_username', 'api_key_secret')
        >>> app_label = pce.labels.create({'key': 'app', 'value': 'App'})
        >>> env_label = pce.labels.create({'key': 'env', 'value': 'Production'})
        >>> loc_label = pce.labels.create({'key': 'loc', 'value': 'AWS'})
        >>> ruleset = RuleSet(
        ...     name='RS-RINGFENCE',
        ...     scopes=[
        ...         LabelSet(
        ...             labels=[app_label, env_label, loc_label]
        ...         )
        ...     ]
        ... )
        >>> ruleset = pce.rule_sets.create(ruleset)
        >>> ruleset
        Ruleset(
            href='/orgs/1/sec_policy/draft/rule_sets/19',
            name='RS-RINGFENCE'
        )
    """
    enabled: bool = None
    scopes: List[LabelSet] = None
    rules: List[Rule] = None
    ip_tables_rules: List[IPTablesRule] = None



import pytest

from illumio.policyobjects import (
    VirtualService,
    ServiceBinding,
    ServicePort,
    ServiceAddress
)
from illumio.util import (
    convert_draft_href_to_active,
    convert_active_href_to_draft,
    convert_protocol,
    DRAFT
)

from helpers import random_string


@pytest.fixture
def virtual_service(pce, session_identifier):
    identifier = random_string()
    virtual_service = pce.virtual_services.create(
        {
            'name': '{}-VS-{}'.format(session_identifier, identifier),
            'description': 'Created by illumio python library integration tests',
            'service_ports': [{'port': 137, 'proto': convert_protocol('udp')}]
        }
    )
    yield virtual_service
    pce.virtual_services.delete(virtual_service.href)


@pytest.fixture
def active_virtual_service(pce, session_identifier, env_label, loc_label):
    identifier = random_string()
    virtual_service = pce.virtual_services.create(
        VirtualService(
            name='{}-VS-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            apply_to='host_only',
            service_addresses=[
                ServiceAddress(fqdn='localhost.localdomain')
            ],
            service_ports=[
                ServicePort(port=80, proto='tcp')
            ],
            labels=[env_label, loc_label],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )

    pce.provision_policy_changes(
        change_description='Provision virtual service',
        hrefs=[virtual_service.href]
    )

    # service bindings must be created using an active virtual service
    virtual_service.href = convert_draft_href_to_active(virtual_service.href)

    yield virtual_service

    # to delete an active object, a DELETE call needs to be made against the
    # draft HREF and then provisioned
    virtual_service.href = convert_active_href_to_draft(virtual_service.href)

    pce.virtual_services.delete(virtual_service.href)
    pce.provision_policy_changes(
        change_description='Remove provisioned virtual service',
        hrefs=[virtual_service.href]
    )


@pytest.fixture
def service_binding(pce, active_virtual_service, workload):
    results = pce.service_bindings.create(
        [ServiceBinding(virtual_service=active_virtual_service, workload=workload)]
    )
    service_binding = results['service_bindings'][0]
    yield service_binding
    pce.service_bindings.delete(service_binding.href)


def test_get_by_reference(pce, virtual_service):
    vs = pce.virtual_services.get_by_reference(virtual_service.href)
    assert vs == virtual_service


def test_get_by_partial_name(pce, session_identifier, virtual_service):
    virtual_services = pce.virtual_services.get(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(virtual_services) == 1


def test_get_by_service_port(pce, virtual_service):
    virtual_services = pce.virtual_services.get(params={'service_ports.port': 137}, policy_version=DRAFT)
    assert len(virtual_services) == 1


def test_get_async(pce, session_identifier, virtual_service):
    virtual_services = pce.virtual_services.get_async(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(virtual_services) == 1


def test_update_virtual_service(pce, virtual_service):
    pce.virtual_services.update(
        virtual_service.href,
        {
            'description': 'Integration test update. Add service address.',
            'service_addresses': [{'fqdn': 'localhost.localdomain'}]
        }
    )
    virtual_services = pce.virtual_services.get(params={'service_addresses.fqdn': 'localhost'}, policy_version=DRAFT)
    vs = virtual_services[0]
    assert vs.href == virtual_service.href and vs.service_addresses[0].fqdn == 'localhost.localdomain'


def test_get_service_binding_by_href(pce, service_binding):
    binding = pce.service_bindings.get_by_reference(service_binding.href)
    assert binding.href == service_binding.href


def test_get_service_bindings_by_virtual_service(pce, active_virtual_service, service_binding):
    bindings = pce.service_bindings.get(params={'virtual_service': active_virtual_service.href})
    assert len(bindings) == 1



# -*- coding: utf-8 -*-

"""This module provides classes related to services and service ports.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio.util import (
    JsonObject,
    MutableObject,
    pce_api,
    convert_protocol
)


@dataclass
class BaseService(JsonObject):
    port: int = None
    proto: int = None

    def __post_init__(self):
        if type(self.proto) is str:
            self.proto = convert_protocol(self.proto)
        super().__post_init__()


@dataclass
class ServicePort(BaseService):
    to_port: int = None
    icmp_type: int = None
    icmp_code: int = None
    service_name: str = None
    process_name: str = None
    windows_service_name: str = None
    user_name: str = None


@dataclass
@pce_api('services', is_sec_policy=True)
class Service(MutableObject):
    """Represents a service in the PCE.

    A service can be port-based or process-based (Windows services).

    Each service contains one or more objects defining the port, protocol, and/or
    process name used by an application running on a workload.

    Service objects are used to write rules or enforcement boundaries to allow
    or deny traffic on its defined ports and processes for workloads in the network.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/services.htm

    Usage:
        >>> from illumio import PolicyComputeEngine, Service, ServicePort
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> pce.set_credentials('api_key_username', 'api_key_secret')
        >>> service = Service(
        ...     name='S-HTTP',
        ...     service_ports=[
        ...         ServicePort(port=80, proto='tcp'),
        ...         ServicePort(port=443, proto='tcp')
        ...     ]
        ... )
        >>> service = pce.services.create(service)
        >>> service
        Service(
            href='/orgs/1/sec_policy/draft/services/15',
            name='S-HTTP',
            service_ports=[
                ServicePort(
                    port=80,
                    proto=6,
                    ...
                ),
                ...
            ],
            ...
        )
    """
    service_ports: List[ServicePort] = None



# -*- coding: utf-8 -*-

"""This module provides classes related to iptables rules.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio.util import JsonObject, MutableObject

from .actor import Actor


@dataclass
class IPTablesStatement(JsonObject):
    table_name: str = None
    chain_name: str = None
    parameters: str = None


@dataclass
class IPTablesRule(MutableObject):
    ip_version: str = None
    enabled: bool = None
    statements: List[IPTablesStatement] = None
    actors: List[Actor] = None



# -*- coding: utf-8 -*-

"""
Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .actor import *
from .rule import *
from .iptablesrule import *
from .enforcementboundary import *
from .ruleset import *



# -*- coding: utf-8 -*-

"""
Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .exceptions import *
from .secpolicy import *
from .util import *
from .accessmanagement import *
from .policyobjects import *
from .infrastructure import *
from .vulnerabilities import *
from .workloads import *
from .rules import *
from .explorer import *
from .pce import *



from illumio.util import DRAFT


def test_get_by_reference(pce, rule_set):
    rs = pce.rule_sets.get_by_reference(rule_set.href)
    assert rs == rule_set


def test_get_by_partial_name(pce, session_identifier, rule_set):
    rule_sets = pce.rule_sets.get(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(rule_sets) == 1


def test_get_async(pce, session_identifier, rule_set):
    rule_sets = pce.rule_sets.get_async(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(rule_sets) == 1


def test_update_rule_set(pce, rule_set):
    pce.rule_sets.update(
        rule_set.href,
        {
            'description': 'Integration test update. Disable rule set.',
            'enabled': False
        }
    )
    rs = pce.rule_sets.get_by_reference(rule_set.href)
    assert rs.enabled == False



# -*- coding: utf-8 -*-

"""This module provides classes related to pairing profile objects.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List, Union

from illumio import IllumioException
from illumio.util import (
    Reference,
    MutableObject,
    EnforcementMode,
    VisibilityLevel,
    pce_api
)


@dataclass
@pce_api('pairing_profiles')
class PairingProfile(MutableObject):
    """Represents a pairing profile in the PCE.

    Pairing profiles are used to configure VEN defaults and generate keys
    for VEN pairing  on workloads that will be managed by the PCE.

    Usage:
        >>> from illumio.workloads import PairingProfile
        >>> pairing_profile = PairingProfile(
        ...     name='PP-DATABASE-VENS',
        ...     enabled=True,
        ...     enforcement_mode='visibility_only',
        ...     visibility_level='flows_summary'
        ... )
        >>> pairing_profile = pce.create_pairing_profile(pairing_profile)
        >>> pairing_profile
        PairingProfile(
            href='/orgs/1/pairing_profiles/19',
            name='PP-DATABASE-VENS',
            enabled=True,
            enforcement_mode='visibility_only',
            visibility_level='flows_summary',
            ...
        )
    """
    enabled: bool = None
    agent_software_release: str = None
    enforcement_mode: str = None
    enforcement_mode_lock: bool = None
    visibility_level: str = None
    visibility_level_lock: bool = None
    allowed_uses_per_key: Union[str, int] = None
    key_lifespan: Union[str, int] = None

    labels: List[Reference] = None
    role_label_lock: bool = None
    app_label_lock: bool = None
    env_label_lock: bool = None
    loc_label_lock: bool = None

    total_use_count: int = None
    is_default: bool = None
    last_pairing_at: str = None

    def _validate(self):
        if self.enforcement_mode and not self.enforcement_mode in EnforcementMode:
            raise IllumioException("Invalid enforcement_mode: {}".format(self.enforcement_mode))
        if self.visibility_level and not self.visibility_level in VisibilityLevel:
            raise IllumioException("Invalid visibility_level: {}".format(self.visibility_level))
        super()._validate()



# -*- coding: utf-8 -*-

"""This module provides classes related to Virtual Enfocement Node objects.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass, field
from typing import List

from illumio import IllumioException
from illumio.util import JsonObject, MutableObject, VisibilityLevel, pce_api


@dataclass
class AgentConfig(JsonObject):
    mode: str = None
    log_traffic: bool = None
    security_policy_update_mode: str = None
    visibility_level: str = None

    def _validate(self):
        if self.visibility_level and not self.visibility_level in VisibilityLevel:
            raise IllumioException("Invalid visibility_level: {}".format(self.visibility_level))
        super()._validate()


@dataclass
class SecureConnect(JsonObject):
    matching_issuer_name: str = None


@dataclass
class AgentHealthErrors(JsonObject):
    errors: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)


@dataclass
class AgentHealth(JsonObject):
    type: str = None
    severity: str = None
    audit_event: str = None


@dataclass
class AgentStatus(JsonObject):
    status: str = None
    uid: str = None
    instance_id: str = None
    last_heartbeat_on: str = None
    uptime_seconds: int = None
    agent_version: str = None
    managed_since: str = None
    fw_config_current: bool = None
    firewall_rule_count: int = None
    security_policy_refresh_at: str = None
    security_policy_applied_at: str = None
    security_policy_received_at: str = None
    agent_health_errors: AgentHealthErrors = None
    agent_health: List[AgentHealth] = None
    security_policy_sync_state: str = None


@dataclass
class VENAgent(MutableObject):
    config: AgentConfig = None
    secure_connect: SecureConnect = None
    status: AgentStatus = None
    unpair_allowed: bool = None
    active_pce_fqdn: str = None
    target_pce_fqdn: str = None
    type: str = None


@dataclass
@pce_api('vens')
class VEN(MutableObject):
    hostname: str = None
    name: str = None
    status: str = None



import pytest

from illumio.policyobjects import LabelGroup

from helpers import random_string


@pytest.fixture
def label_group(pce, session_identifier, role_label):
    identifier = random_string()
    label_group = pce.label_groups.create(
        LabelGroup(
            key='role',
            name='{}-LG-R-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            labels=[role_label],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label_group
    pce.label_groups.delete(label_group.href)


def test_get_by_reference(pce, label_group):
    lg = pce.label_groups.get_by_reference(label_group.href)
    assert lg == label_group


def test_get_by_key(pce, session_identifier, label_group):
    label_groups = pce.label_groups.get(params={'key': 'role', 'name': session_identifier})
    assert len(label_groups) == 1


def test_get_async(pce, session_identifier, label_group):
    label_groups = pce.label_groups.get_async(params={'name': session_identifier})
    assert len(label_groups) == 1


def test_add_label_to_group(pce, session_identifier, label_group, request):
    identifier = random_string()
    role_label = pce.labels.create({
        'key': 'role',
        'value': '{}-R-{}'.format(session_identifier, identifier),
        'external_data_set': session_identifier,
        'external_data_reference': identifier
    })

    def _teardown():
        pce.label_groups.update(label_group.href, {'labels': []})
        pce.labels.delete(role_label.href)

    request.addfinalizer(_teardown)

    label_group.labels.append(role_label)
    pce.label_groups.update(label_group.href, {
        'labels': label_group.labels
    })

    lg = pce.label_groups.get_by_reference(label_group.href)
    assert len(lg.labels) == 2



# -*- coding: utf-8 -*-

"""Stub for PCE interface function definitions"""
from typing import Any, List, overload

from requests import Response

from illumio import (
    IllumioObject,
    Reference,
    IPList,
    TrafficQuery,
    TrafficFlow,
    PolicyVersion
)


class PolicyComputeEngine:

    base_url: str
    org_id: str

    def __init__(self, url: str, port: str, version: str, org_id: str) -> None: ...

    def _setup_retry(self): ...

    def set_credentials(self, username: str, password: str) -> None: ...

    def set_proxies(self, http_proxy: str, https_proxy: str) -> None: ...

    def _request(self, method: str, endpoint: str, include_org: bool, **kwargs) -> Response: ...

    def _build_url(self, endpoint: str, include_org: bool): ...

    def _encode_body(self, kwargs): ...

    def _get_error_message_from_response(self, response: Response) -> str: ...

    def get(self, endpoint: str, **kwargs) -> Response: ...

    def post(self, endpoint: str, **kwargs) -> Response: ...

    def put(self, endpoint: str, **kwargs) -> Response: ...

    def delete(self, endpoint: str, **kwargs) -> Response: ...

    def get_collection(self, endpoint: str, **kwargs) -> Response: ...

    @overload
    def _async_poll(self, job_location: str, retry_time: int) -> str: ...
    @overload
    def _async_poll(self, job_location: str, retry_time: float) -> str: ...

    def check_connection(self, **kwargs) -> bool: ...

    class _PCEObjectAPI:
        def __init__(self, pce: 'PolicyComputeEngine', api_data: object) -> None: ...
        def _build_endpoint(self, policy_version: str, parent: Any) -> str: ...

        @overload
        def get_by_reference(self, reference: str, **kwargs) -> IllumioObject: ...
        @overload
        def get_by_reference(self, reference: Reference, **kwargs) -> IllumioObject: ...
        @overload
        def get_by_reference(self, reference: dict, **kwargs) -> IllumioObject: ...

        @overload
        def get(self, policy_version: str, parent: str, **kwargs) -> List[IllumioObject]: ...
        @overload
        def get(self, policy_version: str, parent: Reference, **kwargs) -> List[IllumioObject]: ...
        @overload
        def get(self, policy_version: str, parent: dict, **kwargs) -> List[IllumioObject]: ...

        @overload
        def get_all(self, policy_version: str, parent: str, **kwargs) -> List[IllumioObject]: ...
        @overload
        def get_all(self, policy_version: str, parent: Reference, **kwargs) -> List[IllumioObject]: ...
        @overload
        def get_all(self, policy_version: str, parent: dict, **kwargs) -> List[IllumioObject]: ...

        @overload
        def get_async(self, policy_version: str, parent: str, **kwargs) -> List[IllumioObject]: ...
        @overload
        def get_async(self, policy_version: str, parent: Reference, **kwargs) -> List[IllumioObject]: ...
        @overload
        def get_async(self, policy_version: str, parent: dict, **kwargs) -> List[IllumioObject]: ...

        @overload
        def create(self, body: Any, parent: str, **kwargs) -> IllumioObject: ...
        @overload
        def create(self, body: Any, parent: Reference, **kwargs) -> IllumioObject: ...
        @overload
        def create(self, body: Any, parent: dict, **kwargs) -> IllumioObject: ...

        @overload
        def update(self, reference: str, body: Any, **kwargs) -> None: ...
        @overload
        def update(self, reference: Reference, body: Any, **kwargs) -> None: ...
        @overload
        def update(self, reference: dict, body: Any, **kwargs) -> None: ...

        @overload
        def delete(self, reference: str, **kwargs) -> None: ...
        @overload
        def delete(self, reference: Reference, **kwargs) -> None: ...
        @overload
        def delete(self, reference: dict, **kwargs) -> None: ...

        def bulk_create(self, objects_to_create: List[IllumioObject], **kwargs) -> List[dict]: ...
        def bulk_update(self, objects_to_update: List[IllumioObject], **kwargs) -> List[dict]: ...

        @overload
        def bulk_delete(self, refs: List[str], **kwargs) -> List[dict]: ...
        @overload
        def bulk_delete(self, refs: List[Reference], **kwargs) -> List[dict]: ...
        @overload
        def bulk_delete(self, refs: List[dict], **kwargs) -> List[dict]: ...

    def __getattr__(self, name: str) -> _PCEObjectAPI: ...

    # Dynamic APIs stubbed here for auto-completion
    container_clusters: _PCEObjectAPI
    enforcement_boundaries: _PCEObjectAPI
    ip_lists: _PCEObjectAPI
    label_groups: _PCEObjectAPI
    labels: _PCEObjectAPI
    pairing_profiles: _PCEObjectAPI
    rule_sets: _PCEObjectAPI
    sec_rules: _PCEObjectAPI
    security_principals: _PCEObjectAPI
    service_bindings: _PCEObjectAPI
    services: _PCEObjectAPI
    vens: _PCEObjectAPI
    virtual_services: _PCEObjectAPI
    workloads: _PCEObjectAPI

    def get_default_ip_list(self, **kwargs) -> IPList: ...

    def generate_pairing_key(self, pairing_profile_href: str, **kwargs) -> str: ...

    def get_traffic_flows(self, traffic_query: TrafficQuery, **kwargs) -> List[TrafficFlow]: ...

    def get_traffic_flows_async(self, query_name: str, traffic_query: TrafficQuery, **kwargs) -> List[TrafficFlow]: ...

    def provision_policy_changes(self, change_description: str, hrefs: List[str], **kwargs) -> PolicyVersion: ...



import json
import re
from copy import copy
from random import randint
from typing import Any
from urllib.parse import unquote_plus
from uuid import uuid4

from illumio.util import ACTIVE, DRAFT


def _gen_uuid():
    return uuid4()


def _gen_sid():
    # see https://docs.microsoft.com/en-US/windows/security/identity-protection/access-control/security-identifiers
    Y = '-'.join([str(randint(1, 9999999999)) for _ in range(randint(1, 5))])
    return 'S-1-{}-{}-{}'.format(randint(1, 5), randint(1, 99), Y)


def _id_seq_generator():
    next_id = 500
    while True:
        yield next_id
        next_id += 1

_id_sequence = _id_seq_generator()


def _gen_int_id():
    return next(_id_sequence)

OBJECT_TYPE_REF_MAP = {
    'container_clusters': _gen_uuid,
    'enforcement_boundaries': _gen_int_id,
    'ip_lists': _gen_int_id,
    'label_groups': _gen_uuid,
    'labels': _gen_int_id,
    'pairing_profiles': _gen_int_id,
    'rule_sets': _gen_int_id,
    'sec_rules': _gen_int_id,
    'security_principals': _gen_sid,
    'service_bindings': _gen_int_id,
    'services': _gen_int_id,
    'vens': _gen_uuid,
    'virtual_services': _gen_uuid,
    'workloads': _gen_uuid
}


class PCEObjectMock(object):
    """
    Base class for PCE object mocks
    """
    base_pattern = '^/api/v2/orgs/\d+(?:/sec_policy/(?:draft|active))?/([a-zA-Z_\-]+)'
    href_pattern = '^/api/v2(/orgs/\d+(?:/sec_policy/(?:draft|active))?(?:/[a-zA-Z_\-]+/[a-zA-Z0-9\-]+)+)$'

    def __init__(self) -> None:
        self.mock_objects = []

    def add_mock_objects(self, mock_objects):
        self.mock_objects += mock_objects

    def get_mock_objects(self, path) -> Any:
        pattern = re.compile(self.base_pattern)
        match = re.match(pattern, path)
        if not match:
            raise Exception("Invalid path: {}".format(path))
        json = self._get_policy_object_by_href(path)
        if json is None:
            json = self._get_matching_objects(path)
        return json

    def _get_policy_object_by_href(self, path):
        href_capture_pattern = re.compile(self.href_pattern)
        match = re.match(href_capture_pattern, path)
        if match:
            for o in self.mock_objects:
                if match.group(1) == o['href']:
                    return o
            return {}
        return None

    def _get_matching_objects(self, path):
        matching_objects = []
        for o in self._object_sieve(path):
            match = True
            query_pattern = re.compile('([a-zA-Z0-9_\-\+%]+)=([a-zA-Z0-9_\-\+%\.\\\/~]+)')
            for param_match in re.finditer(query_pattern, path):
                key, value = param_match.group(1), unquote_plus(param_match.group(2), encoding='utf-8')
                if key not in o:
                    continue
                if key == 'labels':
                    # labels param is a list of lists of label HREFs
                    # example: [["/orgs/1/labels/1", "/orgs/1/labels/2"], ["/orgs/1/labels/3"]]
                    # is equivalent to (1 AND 2) OR 3
                    label_hrefs = [label['href'] for label in o[key]]
                    label_match = False
                    for label_set in json.loads(value):
                        label_set_match = True
                        for label_href in label_set:
                            if label_href not in label_hrefs:
                                label_set_match = False
                        if label_set_match:
                            label_match = True
                            break
                    if not label_match:
                        match = False
                elif value not in o[key]:
                    match = False
            if match:
                matching_objects.append(o)
        return matching_objects

    def _object_sieve(self, path):
        if '/sec_policy/' in path:
            if '/{}/'.format(ACTIVE) in path:
                return [o for o in self.mock_objects if ACTIVE in o['href']]
            return [o for o in self.mock_objects if DRAFT in o['href']]
        return self.mock_objects

    def create_mock_object(self, path, body):
        pattern = re.compile(self.base_pattern)
        match = re.match(pattern, path)
        if not match:
            raise Exception("Invalid path: {}".format(path))
        object_type = match.group(1)
        href = '/orgs/{}/{}'.format(path.split('/orgs/')[-1], OBJECT_TYPE_REF_MAP[object_type]())
        body['href'] = href
        self.mock_objects.append(body)
        return body

    def update_mock_object(self, path, body):
        href_capture_pattern = re.compile(self.href_pattern)
        match = re.match(href_capture_pattern, path)
        if match:
            _mock_objects = copy(self.mock_objects)
            for o in _mock_objects:
                if match.group(1) == o['href']:
                    for k, v in body.items():
                        o[k] = v
                    return
            raise Exception("Attempting to update invalid or missing object")
        raise Exception("Invalid HREF passed to update_mock_object")

    def delete_mock_object(self, path):
        href_capture_pattern = re.compile(self.href_pattern)
        match = re.match(href_capture_pattern, path)
        if match:
            _mock_objects = copy(self.mock_objects)
            for o in _mock_objects:
                if match.group(1) == o['href']:
                    self.mock_objects.remove(o)
                    return
            raise Exception("Attempting to delete invalid or missing object")
        raise Exception("Invalid HREF passed to delete_mock_object")



# -*- coding: utf-8 -*-

"""This module provides the core of the class model used in the client.

The base JsonObject class handles basic encoding and decoding to translate
responses from the PCE REST API into python objects. The implementation
leverages dataclasses to simplify the logic for these operations.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import copy
import json
from abc import ABC
from dataclasses import dataclass, fields
from inspect import signature, isclass
from typing import List, Any

from illumio.exceptions import IllumioException

from .constants import IllumioEnumMeta
from .functions import ignore_empty_keys, isunion, islist

_default = json.JSONEncoder()  # fall back to the default encoder for non-Illumio API objects


class IllumioEncoder(json.JSONEncoder):
    """Convenience class for encoding JsonObjects.

    >>> json.dumps(flow, cls=IllumioEncoder, indent=4)
    """
    def default(self, o: Any) -> Any:
        return getattr(o.__class__, "to_json", _default.default)(o)


@dataclass
class JsonObject(ABC):

    def __post_init__(self):
        for field in fields(self):
            value = getattr(self, field.name)
            self._flatten_ref(field, value)
            self._resolve_enum(field, value)
        self._validate()

    def _resolve_enum(self, field, value):
        """Replaces IllumioEnumMeta subtypes with their internal value.

        This allows clients to pass enums directly as attribute values.

        For example:

        >>> Workload(..., enforcement_mode=EnforcementMode.SELECTIVE)

        will be converted to

        >>> Workload(..., enforcement_mode='selective')
        """
        if value is None:
            return
        if isinstance(type(value), IllumioEnumMeta):
            setattr(self, field.name, value.value)

    def _flatten_ref(self, field, value):
        """Replaces Reference subclasses with a simplified Reference object.

        This allows clients to pass a Reference subclass instance without
        breaking the encoded object schema for API calls.
        """
        if value is None:
            return
        if field.type is Reference:
            if isinstance(value, Reference):
                setattr(self, field.name, Reference(value.href))
        elif islist(field.type):
            if field.type.__args__[0] is Reference:
                ref_list = []
                for ref in value:
                    if isinstance(ref, Reference):
                        ref_list.append(Reference(href=ref.href))
                    else:
                        ref_list.append(ref)
                setattr(self, field.name, ref_list)
        elif isunion(field.type):
            if Reference in field.type.__args__:
                if isinstance(value, Reference):
                    setattr(self, field.name, Reference(value.href))

    def _validate(self):
        """Validates fields by comparing their values to their registered dataclass types."""
        for field in fields(self):
            value = getattr(self, field.name)
            if not self._validate_field(field.type, value):
                raise AttributeError("Invalid value for {}: {}. Must be of type {}".format(field.name, value, field.type))

    def _validate_field(self, expected_type, value) -> bool:
        if value is None:
            return True
        elif expected_type is object:
            return True
        elif type(value) == expected_type:
            return True
        elif isunion(expected_type):
            return any(self._validate_field(type_, value) for type_ in expected_type.__args__)
        elif isclass(expected_type) and issubclass(expected_type, JsonObject):
            # if the object is already decoded, determine whether it's
            # a subtype of what the field expects
            if isinstance(value, JsonObject):
                return isinstance(value, expected_type)
            # XXX: otherwise, expand objects to run their own validation.
            #   this is *slow*, but needed to validate deeply nested types
            expected_type.from_json(value)
            return True
        elif islist(expected_type) and isinstance(value, list):
            if not value:
                return True  # empty lists are always valid
            expected_type = expected_type.__args__[0]
            return all(self._validate_field(expected_type, o) for o in value)
        return False

    def to_json(self) -> Any:
        return deep_encode(self)

    @classmethod
    def from_json(cls, data: Any) -> 'JsonObject':
        """
        Given a JSON object or dictionary, decodes the data as an object
        of the calling JsonObject subtype. Accepts arbitrary key/value pairs
        as a form of forwards-compatibility.

        Classes can optionally extend decoding with the _decode_complex_types
        function for complex-type members.

        Based in part on https://stackoverflow.com/a/55101438
        """
        data = json.loads(data) if type(data) is str else data
        cls_fields = {field for field in signature(cls).parameters}

        defined_params, undefined_params = {}, {}
        for k, v in data.items():
            if k in cls_fields:
                defined_params[k] = v
            else:
                undefined_params[k] = v

        o = cls(**defined_params)

        for k, v in undefined_params.items():
            setattr(o, k, v)

        o._decode_complex_types()
        return o

    def _decode_complex_types(self) -> None:
        for field in fields(self):
            value = getattr(self, field.name)
            decoded_value = self._decode_field(field.type, value)
            setattr(self, field.name, decoded_value)

    def _decode_field(self, type_, value) -> Any:
        if value is None:
            return None
        if isinstance(value, JsonObject):
            # if the value has already been decoded, return it
            return value
        elif isclass(type_) and issubclass(type_, JsonObject):
            return type_.from_json(value)
        elif isinstance(value, list):
            # if the value is a list, expect the field type to be List[T]
            type_ = type_.__args__[0]
            return list(self._decode_field(type_, o) for o in value)
        return value


def deep_encode(o: Any) -> Any:
    """
    Recursively encode members of the given object and return a JSON-compatible
    copy. Children of the JsonObject superclass can optionally implement an
    _encode method to provide a customized encoding response, otherwise the
    default ignore_empty_keys function is called to remove null value pairs.

    Functionally similar to the dataclasses asdict method, but with the necessary
    adjustment of calling an optional custom encoding function for types that
    don't strictly mirror their dataclass field pairs when encoded.
    """
    if isinstance(o, JsonObject):
        result = []
        if hasattr(type(o), '_encode'):
            return o._encode()
        for f in fields(o):
            value = deep_encode(getattr(o, f.name))
            result.append((f.name, value))
        return ignore_empty_keys(result)
    elif isinstance(o, (list, tuple)):
        return type(o)(deep_encode(o) for o in o)
    elif isinstance(o, dict):
        return type(o)((deep_encode(k), deep_encode(v)) for k, v in o.items())
    else:
        return copy.deepcopy(o)


@dataclass
class Reference(JsonObject):
    href: str = None


def href_from(reference: Any):
    """Attempts to parse HREF value from a provided source."""
    if isinstance(reference, Reference):
        return reference.href
    elif type(reference) is dict:
        if 'href' in reference:
            return reference['href']
    elif type(reference) is str:
        return reference
    raise IllumioException('Failed to extract HREF from value: {}'.format(reference))


@dataclass
class IllumioObject(Reference):
    name: str = None
    description: str = None
    external_data_set: str = None
    external_data_reference: str = None


@dataclass
class MutableObject(IllumioObject):
    created_at: str = None
    updated_at: str = None
    deleted_at: str = None
    update_type: str = None
    delete_type: str = None
    created_by: Reference = None
    updated_by: Reference = None
    deleted_by: Reference = None
    caps: List[str] = None


@dataclass
class ImmutableObject(IllumioObject):
    created_at: str = None
    created_by: Reference = None



import pytest

from illumio.policyobjects import ServicePort
from illumio.rules import EnforcementBoundary, AMS
from illumio.util import convert_protocol, DRAFT

from helpers import random_string


@pytest.fixture
def enforcement_boundary(pce, session_identifier, rdp_service):
    identifier = random_string()
    hostname = '{}.{}'.format(session_identifier, identifier)
    any_enforcement_boundary = pce.get_default_ip_list()
    enforcement_boundary = pce.enforcement_boundaries.create(
        EnforcementBoundary.build(
            name=hostname,
            ingress_services=[rdp_service, {'port': 20009, 'proto': convert_protocol('tcp')}],
            providers=[AMS],
            consumers=[any_enforcement_boundary]
        )
    )
    yield enforcement_boundary
    pce.enforcement_boundaries.delete(enforcement_boundary.href)


def test_get_by_reference(pce, enforcement_boundary):
    eb = pce.enforcement_boundaries.get_by_reference(enforcement_boundary.href)
    assert eb == enforcement_boundary


def test_get_by_partial_name(pce, session_identifier, enforcement_boundary):
    enforcement_boundaries = pce.enforcement_boundaries.get(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(enforcement_boundaries) == 1


def test_get_async(pce, session_identifier, enforcement_boundary):
    enforcement_boundaries = pce.enforcement_boundaries.get_async(params={'name': session_identifier}, policy_version=DRAFT)
    assert len(enforcement_boundaries) == 1


def test_update_enforcement_boundary(pce, enforcement_boundary):
    services = enforcement_boundary.ingress_services
    services.append(ServicePort(port=20009, proto='udp'))
    pce.enforcement_boundaries.update(
        enforcement_boundary.href, {'ingress_services': services}
    )
    eb = pce.enforcement_boundaries.get_by_reference(enforcement_boundary.href)
    assert len(eb.ingress_services) == 3



# -*- coding: utf-8 -*-

"""This module provides classes related to virtual services.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio import IllumioException
from illumio.util import IllumioObject, Reference, MutableObject, pce_api

from .service import BaseService, ServicePort

HOST_ONLY = 'host_only'
INTERNAL_BRIDGE_NETWORK = 'internal_bridge_network'


@dataclass
class ServiceAddress(BaseService):
    """Defines a Service Address record for use in a Virtual Service address pool.

    A ServiceAddress can be defined with either an IP address or FQDN.

    If an IP is provided, one of a port number or network object HREF
    must be provided as well.

    If an FQDN is given it's sufficient by itself, but description or
    port values can optionally be provided.
    """
    ip: str = None
    fqdn: str = None
    network: Reference = None
    description: str = None


@dataclass
@pce_api('virtual_services', is_sec_policy=True)
class VirtualService(MutableObject):
    apply_to: str = None
    pce_fqdn: str = None
    service: Reference = None
    service_ports: List[ServicePort] = None
    service_addresses: List[ServiceAddress] = None
    ip_overrides: List[str] = None
    labels: List[Reference] = None

    def _validate(self):
        if self.apply_to and self.apply_to not in {HOST_ONLY, INTERNAL_BRIDGE_NETWORK}:
            raise IllumioException("Invalid 'apply_to' value: {}".format(self.apply_to))
        super()._validate()


@dataclass
class PortOverride(BaseService):
    to_port: int = None
    new_port: int = None
    new_to_port: int = None


@dataclass
@pce_api('service_bindings')
class ServiceBinding(IllumioObject):
    virtual_service: Reference = None
    workload: Reference = None
    port_overrides: List[PortOverride] = None



import json
import os
import re
from typing import List

import pytest
from requests_mock import ANY

from illumio.rules import Rule

MOCK_RULES = os.path.join(pytest.DATA_DIR, 'rules.json')
MOCK_RULE_SET_HREF = '/orgs/1/sec_policy/draft/rule_sets/1'


@pytest.fixture(scope='module')
def rules() -> List[Rule]:
    with open(MOCK_RULES, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(autouse=True)
def rules_mock(pce_object_mock, rules):
    pce_object_mock.add_mock_objects(rules)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    pattern = re.compile('/sec_rules')
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)


@pytest.fixture()
def mock_rule(pce) -> Rule:
    yield pce.rules.get_by_reference('{}/sec_rules/1'.format(MOCK_RULE_SET_HREF))


def test_label_resolution_block(mock_rule):
    json_rule = mock_rule.to_json()
    assert json_rule['resolve_labels_as']['providers'] == ['workloads']


def test_builder():
    rule = Rule.build(
        providers=['/orgs/1/labels/1'],
        consumers=['ams'],
        ingress_services=[{'port': 1234, 'proto': 6}],
        resolve_providers_as=['workloads'],
        resolve_consumers_as=['workloads']
    )
    expected_result = json.loads('''
        {
            "enabled": true,
            "ingress_services": [
                {"port": 1234, "proto": 6}
            ],
            "providers": [
                {"label": {"href": "/orgs/1/labels/1"}}
            ],
            "consumers": [
                {"actors": "ams"}
            ],
            "resolve_labels_as": {
                "providers": ["workloads"],
                "consumers": ["workloads"]
            }
        }
    ''')
    assert rule.to_json() == expected_result


def test_get_rules(pce):
    rules = pce.rules.get(parent=MOCK_RULE_SET_HREF)
    assert len(rules) > 0



from illumio.workloads import Workload

from helpers import random_string


def test_get_by_reference(pce, workload):
    wl = pce.workloads.get_by_reference(workload.href)
    assert wl == workload


def test_get_by_partial_name(pce, session_identifier, workload):
    workloads = pce.workloads.get(params={'name': session_identifier})
    assert len(workloads) == 1


def test_get_async(pce, session_identifier, workload):
    workloads = pce.workloads.get_async(params={'name': session_identifier})
    assert len(workloads) == 1


def test_update_workload(pce, workload):
    pce.workloads.update(
        workload.href,
        {
            'public_ip': '127.0.0.1',
            'description': 'Integration test update. Disable rule set.'
        }
    )
    wl = pce.workloads.get_by_reference(workload.href)
    assert wl.public_ip == '127.0.0.1'


def test_bulk_create(pce, session_identifier, request):
    id_1 = random_string()
    id_2 = random_string()

    hostname_1 = '{}.{}'.format(session_identifier, id_1)
    hostname_2 = '{}.{}'.format(session_identifier, id_2)

    workloads = [
        Workload(
            name=hostname_1,
            hostname=hostname_1,
            # TODO
            description='Created by illumio python library integration tests',
            external_data_set=session_identifier,
            external_data_reference=id_1
        ),
        {
            'name': hostname_2,
            'hostname': hostname_2,
            'description': 'Created by illumio python library integration tests',
            'external_data_set': session_identifier,
            'external_data_reference': id_2
        }
    ]

    results = pce.workloads.bulk_create(workloads)

    def _teardown():
        pce.workloads.bulk_delete([result['href'] for result in results])

    request.addfinalizer(_teardown)


    for result in results:
        assert not result['errors']
    workloads = pce.workloads.get(params={'external_data_set': session_identifier})
    assert len(workloads) == 2


def test_bulk_update(pce, session_identifier, workload, request):
    identifier = random_string()
    hostname = '{}.{}'.format(session_identifier, identifier)

    new_workload = pce.workloads.create(
        Workload(
            name=hostname,
            hostname=hostname,
            description='Created by illumio python library integration tests',
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )

    def _teardown():
        pce.workloads.delete(new_workload)

    request.addfinalizer(_teardown)

    workloads = [
        {
            'href': wl.href,
            'enforcement_mode': 'selective'
        } for wl in [workload, new_workload]
    ]

    results = pce.workloads.bulk_update(workloads)
    for result in results:
        assert not result['errors']
    workloads = pce.workloads.get(params={'external_data_set': session_identifier})
    for wl in workloads:
        assert wl.enforcement_mode == 'selective'


def test_bulk_delete(pce, session_identifier):
    id_1 = random_string()
    id_2 = random_string()

    hostname_1 = '{}.{}'.format(session_identifier, id_1)
    hostname_2 = '{}.{}'.format(session_identifier, id_2)

    workloads = [
        Workload(
            name=hostname_1,
            hostname=hostname_1,
            # TODO
            description='Created by illumio python library integration tests',
            external_data_set=session_identifier,
            external_data_reference=id_1
        ),
        {
            'name': hostname_2,
            'hostname': hostname_2,
            'description': 'Created by illumio python library integration tests',
            'external_data_set': session_identifier,
            'external_data_reference': id_2
        }
    ]

    created_workloads = []

    for workload in workloads:
        created_workloads.append(pce.workloads.create(workload))

    results = pce.workloads.bulk_delete(created_workloads)
    for result in results:
        assert not result['errors']
    workloads = pce.workloads.get(params={'external_data_set': session_identifier})
    assert len(workloads) == 0



import os

import pytest

from illumio import PolicyComputeEngine, IllumioException
from illumio.util import PCE_APIS

from helpers import random_string

# environment variables for integration tests
pce_host = os.getenv('ILLUMIO_PCE_HOST')
pce_port = os.getenv('ILLUMIO_PCE_PORT', 443)
org_id = os.getenv('ILLUMIO_PCE_ORG_ID', 1)
api_key = os.getenv('ILLUMIO_API_KEY_USERNAME')
api_secret = os.getenv('ILLUMIO_API_KEY_SECRET')


@pytest.fixture(scope='session')
def pce():
    if not pce_host or not api_key or not api_secret:
        raise IllumioException('''Missing required environment variable for integration tests.
Make sure ILLUMIO_PCE_HOST, ILLUMIO_API_KEY_USERNAME, and ILLUMIO_API_KEY_SECRET are set.''')
    pce = PolicyComputeEngine(pce_host, port=pce_port, org_id=org_id)
    pce.set_credentials(api_key, api_secret)
    return pce


@pytest.fixture(scope='session')
def session_identifier():
    return 'illumio-py-integration-{}'.format(random_string())


@pytest.fixture(scope='session', autouse=True)
def cleanup(pce, session_identifier, request):
    def _teardown():
        for name in PCE_APIS.keys():
            api = getattr(pce, name)
            try:
                leftover_integration_objects = api.get(
                    params={'external_data_set': session_identifier}
                )
                for o in leftover_integration_objects:
                    api.delete(o.href)
            except IllumioException:
                pass
    request.addfinalizer(_teardown)


# prepare common fixtures
from common import *



import json
import os
import re
from typing import List

import pytest

from illumio import IllumioException
from illumio.policyobjects import VirtualService, ServicePort, ServiceAddress
from illumio.util import Reference, DRAFT

VIRTUAL_SERVICES = os.path.join(pytest.DATA_DIR, 'virtual_services.json')


@pytest.fixture(scope='module')
def virtual_services() -> List[dict]:
    with open(VIRTUAL_SERVICES, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(scope='module')
def new_virtual_service() -> VirtualService:
    return VirtualService(
        name="VS-TEST",
        service_ports=[
            ServicePort(port=443, proto="TCP")
        ]
    )


@pytest.fixture(autouse=True)
def virtual_services_mock(pce_object_mock, virtual_services):
    pce_object_mock.add_mock_objects(virtual_services)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    pattern = re.compile('/sec_policy/(draft|active)/virtual_services')
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)


@pytest.fixture()
def mock_virtual_service(pce):
    yield pce.virtual_services.get_by_reference("/orgs/1/sec_policy/draft/virtual_services/14d7ff69-2fa4-458b-a299-e3f11ffa9b01")


def test_decoded_service_ports(mock_virtual_service: VirtualService):
    assert type(mock_virtual_service.service_ports[0]) is ServicePort


def test_decoded_service_addresses(mock_virtual_service: VirtualService):
    assert type(mock_virtual_service.service_addresses[0]) is ServiceAddress


def test_decoded_labels(mock_virtual_service: VirtualService):
    assert type(mock_virtual_service.labels[0]) is Reference


def test_create_virtual_service(pce, new_virtual_service: VirtualService):
    created_virtual_service = pce.create_virtual_service(new_virtual_service)
    assert created_virtual_service.name == 'VS-TEST'


def test_invalid_protocol_name():
    with pytest.raises(IllumioException):
        VirtualService(
            name='VS-TEST', service_ports=[
                ServicePort(port=443, proto="invalidproto")
            ]
        )


def test_invalid_apply_to_value():
    with pytest.raises(IllumioException):
        VirtualService(href='/test/href', name='VS-TEST', apply_to='invalid_value')


def test_get_by_name(pce):
    virtual_services = pce.virtual_services.get(params={'name': 'VS-'})
    assert len(virtual_services) == 1


def test_get_draft_virtual_services(pce, mock_virtual_service):
    virtual_service = pce.virtual_services.get(params={'name': 'VS-INTERNAL', 'max_results': 1}, policy_version=DRAFT)[0]
    assert virtual_service == mock_virtual_service


def test_create_virtual_service(pce, new_virtual_service):
    created_virtual_service = pce.virtual_services.create(new_virtual_service)
    assert created_virtual_service.href != ''
    virtual_service = pce.virtual_services.get_by_reference(created_virtual_service.href)
    assert created_virtual_service == virtual_service


def test_update_virtual_service(pce, mock_virtual_service):
    pce.virtual_services.update(mock_virtual_service.href, {'enabled': False})
    updated_virtual_service = pce.virtual_services.get_by_reference(mock_virtual_service.href)
    assert updated_virtual_service.enabled is False



# -*- coding: utf-8 -*-

"""This module provides helper functions and decorators for common use-cases.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import functools
import re
import socket
import typing
import warnings
from dataclasses import dataclass
from urllib.parse import urlparse

from illumio._version import version
from illumio.exceptions import IllumioException
from .constants import ACTIVE, DRAFT, PCE_APIS


def ignore_empty_keys(o: dict):
    """Removes keys with None-type values from the provided dict.

    Used for JSON encoding to avoid schema errors due to empty or invalid parameters.
    """
    return {k: v for (k, v) in o if v is not None}


def convert_draft_href_to_active(href: str) -> str:
    return href.replace('/{}/'.format(DRAFT), '/{}/'.format(ACTIVE))


def convert_active_href_to_draft(href: str) -> str:
    return href.replace('/{}/'.format(ACTIVE), '/{}/'.format(DRAFT))


def deprecated(deprecated_in, message=None):
    """
    Deprecation decorator, adapted from https://stackoverflow.com/a/30253848
    Will emit a warning when the decorated function is called.
    """
    def _deprecated(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            default_message = "Call to deprecated function {}. Deprecated in version {}, current version is {}.".format(
                func.__name__, deprecated_in, version
            )
            warning_message = message or default_message
            warnings.warn(warning_message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return _deprecated


def pce_api(name: str, endpoint: str = None, is_sec_policy=False, is_global=False):
    """Decorates an IllumioObject subclass to denote it as a PCE API object type.

    This registers the type in the PCE_APIS mapping used to determine whether
    a given name corresponds to an API-accessible type.

    We can then leverage __getattr__ to instantiate a generic API interface for
    any registered type (with some caveats, see the _PCEObjectAPI documentation).

    By default, registers the endpoint as /{name}, but the endpoint can also be
    specified in the decorator function call.

    For example:

    >>> @pce_api('labels', endpoint='/labels')
    >>> class Label(IllumioObject):
    ...     ...
    >>> pce = PolicyComputeEngine(...)
    >>> # the 'labels' name is registered, and so we can
    >>> # call /labels endpoints through the _PCEObjectAPI interface
    >>> labels = pce.labels.get()
    >>> labels
    [
        Label(
            href='/orgs/1/labels/1',
            key='role',
            value='R-DB',
            ...
        ),
        ...
    ]

    Args:
        name (str): the name of the API. used as a PolicyComputeEngine attribute name
            to generate the API interface.
        endpoint (str, optional): _description_. Defaults to None.
        is_sec_policy (bool, optional): whether or not the object reflects a security
            policy API with the sec_policy/{pversion} prefix. Defaults to False.
        is_global (bool, optional): whether or not the object reflects a global API,
            such as /health or /users. These APIs operate on the entire PCE rather
            than a single tenant, and don't need the /orgs/{org_id} prefix.
    """
    def _decorator(cls):
        @dataclass
        class __PCEApi:
            name: str
            endpoint: str
            object_class: object
            is_sec_policy: bool
            is_global: bool
        PCE_APIS[name] = __PCEApi(
            name=name,
            endpoint=endpoint or '/{}'.format(name),
            object_class=cls,
            is_sec_policy=is_sec_policy,
            is_global=is_global
        )
        return cls
    return _decorator


def parse_url(url: str) -> tuple:
    """Parses given URL into its scheme and hostname, stripping port and path."""
    pattern = re.compile('^\w+://')
    if not re.match(pattern, url):
        url = 'https://{}'.format(url)
    parsed = urlparse(url)
    protocol = parsed.scheme
    if protocol not in ('http', 'https'):
        protocol = 'https'     # only support http(s)
    return protocol, parsed.hostname


def convert_protocol(protocol: str) -> int:
    """Given a protocol name, returns the integer ID of that protocol.

    Usage:
        >>> convert_protocol('tcp')
        6

    Args:
        protocol (str): case-insensitive protocol string, e.g. 'tcp', 'UDP'

    Raises:
        IllumioException: if an invalid protocol name is provided.

    Returns:
        int: the integer ID of the given protocol, e.g. 17 for 'udp'
    """
    try:
        return socket.getprotobyname(protocol)
    except:
        raise IllumioException('Invalid protocol name: {}'.format(protocol))


if hasattr(typing, 'get_origin'):
    # python 3.8+ - introduces the get_origin function
    def isunion(type_):
        return typing.get_origin(type_) is typing.Union

    def islist(type_):
        if type_ is list:
            return True
        return typing.get_origin(type_) is list

elif hasattr(typing, '_GenericAlias'):
    # python 3.7 - changes meta types to be based off the
    # _GenericAlias supertype. __extra__ is now __origin__
    def isunion(type_):
        if isinstance(type_, typing._GenericAlias):
            return type_.__origin__ is typing.Union
        return False

    def islist(type_):
        if type_ is list:
            return True
        if isinstance(type_, typing._GenericAlias):
            return type_.__origin__ is list
        return False

else:
    # python 3.6
    def isunion(type_):
        return isinstance(type_, typing._Union)

    def islist(type_):
        if type_ is list:
            return True
        # in 3.6, List's type is GenericMeta, so we
        # instead check the __extra__ param
        if hasattr(type_, '__extra__'):
            return type_.__extra__ is list
        return False



# -*- coding: utf-8 -*-

"""This module provides classes related to workload objects.

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List, Union

from illumio import IllumioException
from illumio.infrastructure import ContainerCluster
from illumio.policyobjects import Label, BaseService, Service, ServicePort
from illumio.vulnerabilities import Vulnerability
from illumio.util import (
    JsonObject,
    Reference,
    MutableObject,
    LinkState,
    EnforcementMode,
    VisibilityLevel,
    pce_api
)

from .ven import VEN, VENAgent


@dataclass
class Interface(JsonObject):
    name: str = None
    link_state: str = None
    address: str = None
    cidr_block: int = None
    default_gateway_address: str = None
    network: Reference = None
    network_detection_mode: str = None
    friendly_name: str = None
    loopback: bool = None

    def _validate(self):
        if self.link_state and not self.link_state in LinkState:
            raise IllumioException("Invalid link_state: {}".format(self.link_state))
        super()._validate()


@dataclass
class WorkloadServicePort(JsonObject):
    port: int = None
    protocol: int = None
    address: str = None
    user: str = None
    package: str = None
    process_name: str = None
    win_service_name: str = None


@dataclass
class WorkloadServices(JsonObject):
    uptime_seconds: int = None
    open_service_ports: List[WorkloadServicePort] = None


@dataclass
class PortWideExposure(JsonObject):
    any: bool = None
    ip_list: bool = None


@dataclass
class VulnerabilitiesSummary(JsonObject):
    num_vulnerabilities: int = None
    vulnerable_port_exposure: int = None
    vulnerable_port_wide_exposure: PortWideExposure = None
    vulnerability_exposure_score: int = None
    vulnerability_score: int = None
    max_vulnerability_score: int = None


@dataclass
class DetectedVulnerability(BaseService):
    ip_address: str = None
    port_exposure: int = None
    port_wide_exposure: PortWideExposure = None
    workload: Reference = None
    vulnerability: Vulnerability = None
    vulnerability_report: Reference = None


@dataclass
class IKEAuthenticationCertificate(JsonObject):
    pass


@dataclass
@pce_api('workloads')
class Workload(MutableObject):
    """Represents a workload in the PCE.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/workloads/_ch-workloads.htm

    Usage:
        >>> from illumio import PolicyComputeEngine, Workload, Interface, EnforcementMode
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> pce.set_credentials('api_key_username', 'api_key_secret')
        >>> role_label = pce.labels.create({'key': 'role', 'value': 'Web'})
        >>> app_label = pce.labels.create({'key': 'app', 'value': 'App'})
        >>> env_label = pce.labels.create({'key': 'env', 'value': 'Development'})
        >>> loc_label = pce.labels.create({'key': 'loc', 'value': 'NYC-Datacenter'})
        >>> workload = Workload(
        ...     name='Web 01',
        ...     hostname='web01.lab.company.com',
        ...     public_ip='10.8.17.229',
        ...     labels=[role_label, app_label, env_label, loc_label],
        ...     interfaces=[
        ...         Interface(
        ...             name='lo0',
        ...             address='127.0.0.1',
        ...             link_state='up'
        ...         )
        ...     ],
        ...     enforcement_mode=EnforcementMode.SELECTIVE,
        ...     online=True
        ... )
        >>> workload = pce.workloads.create(workload)
        >>> workload
        Workload(
            href='/orgs/1/workloads/572eb23e-a891-42b5-b488-cd9ffe3622f5',
            name='Web 01',
            hostname='web01.lab.company.com',
            public_ip='10.8.17.229',
            labels=[
                Label(key='role', value='Web', ...),
                ...
            ],
            interfaces=[
                Interface(
                    name='lo0',
                    address='127.0.0.1',
                    link_state='up'
                )
            ],
            enforcement_mode='selective',
            online=True,
            ...
        )
    """
    hostname: str = None
    os_type: str = None
    service_principal_name: str = None
    agent_to_pce_certificate_authentication_id: str = None
    distinguished_name: str = None
    public_ip: str = None
    interfaces: List[Interface] = None
    service_provider: str = None
    data_center: str = None
    data_center_zone: str = None
    os_id: str = None
    os_detail: str = None
    online: bool = None
    deleted: bool = None
    ignored_interface_names: List[str] = None
    firewall_coexistence: str = None
    containers_inherit_host_policy: bool = None
    blocked_connection_action: str = None
    labels: List[Reference] = None
    services: WorkloadServices = None
    vulnerabilities_summary: VulnerabilitiesSummary = None
    detected_vulnerabilities: List[DetectedVulnerability] = None
    agent: VENAgent = None
    ven: VEN = None
    enforcement_mode: str = None
    visibility_level: str = None
    num_enforcement_boundaries: int = None
    selectively_enforced_services: List[Union[Service, ServicePort]] = None
    container_cluster: ContainerCluster = None
    ike_authentication_certificate: IKEAuthenticationCertificate = None

    def _validate(self):
        if self.enforcement_mode and not self.enforcement_mode in EnforcementMode:
            raise IllumioException("Invalid enforcement_mode: {}".format(self.enforcement_mode))
        if self.visibility_level and not self.visibility_level in VisibilityLevel:
            raise IllumioException("Invalid visibility_level: {}".format(self.visibility_level))
        super()._validate()

    def _decode_complex_types(self):
        enforced_services = []
        for service in self.selectively_enforced_services or []:
            service_type = Service if 'href' in service else ServicePort
            enforced_services.append(service_type.from_json(service))
        self.selectively_enforced_services = enforced_services if enforced_services else None
        super()._decode_complex_types()



# -*- coding: utf-8 -*-

"""This module provides the core PolicyComputeEngine class for communicating with the PCE.

Usage:
    >>> from illumio import PolicyComputeEngine
    >>> pce = PolicyComputeEngine('my.pce.com', port='8443', org_id='12')
    >>> pce.set_credentials('<API_KEY>', '<API_SECRET>')
    >>> workloads = pce.workloads.get(
    ...     params={
    ...         'managed': True,
    ...         'enforcement_mode': 'visibility_only'
    ...     }
    ... )
    >>> workloads
    [
        Workload(href='/orgs/12/workloads/c754a713-2bde-4427-af1f-bff145be509b', ...),
        ...
    ]

Copyright:
    (c) 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import json
import time
from typing import Any, List, Union

from requests import Session, Response
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .secpolicy import PolicyChangeset, PolicyVersion
from .exceptions import IllumioApiException
from .policyobjects import IPList
from .explorer import TrafficQuery, TrafficFlow
from .util import (
    deprecated,
    convert_active_href_to_draft,
    parse_url,
    href_from,
    Reference,
    IllumioObject,
    IllumioEncoder,
    ACTIVE,
    DRAFT,
    ANY_IP_LIST_NAME,
    BULK_CHANGE_LIMIT,
    PCE_APIS
)


class PolicyComputeEngine:
    """The REST client core for the Illumio Policy Compute Engine.

    Contains request logic for API calls and handles the HTTP(S) connection to the PCE.

    Attributes:
        base_url: the base URL for API calls to the PCE. Has the form
            http[s]://<DOMAIN_NAME>:<PORT>/api/<API_VERSION>
        org_id: the PCE organization ID.
        container_clusters: Container Clusters API.
        enforcement_boundaries: Enforcement Boundaries API.
        ip_lists: IP Lists API.
        label_groups: Label Groups API.
        labels: Labels API.
        pairing_profiles: Pairing Profiles API.
        rule_sets: Rule Sets API.
        sec_rules: Rules API.
        security_principals: Security Principals API.
        service_bindings: Service Bindings API.
        services: Services API.
        vens: VENs API.
        virtual_services: Virtual Services API.
        workloads: Workloads API.
    """
    def __init__(self, url: str, port: str = '443', version: str = 'v2', org_id: str = '1') -> None:
        """Initializes the PCE REST client.

        Args:
            url (str): PCE URL. May include http:// or https:// as the protocol.
            port (str, optional): PCE http(s) port. Defaults to '443'.
            version (str, optional): The PCE API version to use. Defaults to 'v2'.
            org_id (str, optional): The PCE organization ID. Defaults to '1'.
        """
        self._apis = {}
        self._encoder = IllumioEncoder()
        self._session = Session()
        self._session.headers.update({'Accept': 'application/json'})
        protocol, url = parse_url(url)
        self.base_url = "{}://{}:{}/api/{}".format(protocol, url, port, version)
        self.org_id = org_id
        self._setup_retry()

    def _setup_retry(self):
        """Configures `requests.Session` retry defaults"""
        retry_strategy = Retry(
            total=5,  # retry up to 5 times
            # {backoff} * (2 ** ({retry count} - 1))
            # 1, 2, 4, 8, 16 (seconds)
            backoff_factor=2,
            status_forcelist=[429, 500, 502, 503, 504]
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self._session.mount("https://", adapter)
        self._session.mount("http://", adapter)

    def set_credentials(self, username: str, password: str) -> None:
        """Sets username and password values to authenticate with the PCE.

        The provided credentials can be either user credentials or an API key/secret pair.

        Args:
            username (str): username or API key.
            password (str): password or API secret.
        """
        self._session.auth = (username, password)

    def set_proxies(self, http_proxy: str = None, https_proxy: str = None) -> None:
        """Sets proxy information in the request session.

        Args:
            http_proxy (str, optional): HTTP proxy URI. Defaults to None.
            https_proxy (str, optional): HTTPS proxy URI. Defaults to None.
        """
        self._session.proxies.update({'http': http_proxy, 'https': https_proxy})

    def _request(self, method: str, endpoint: str, include_org=True, **kwargs) -> Response:
        """Makes an API call to the PCE.

        Args:
            method (str): the HTTP request method. Supports the same verbs as `requests.request`.
            endpoint (str): the API endpoint to call.
            include_org (bool, optional): whether or not to include /orgs/{org_id} in the API call.
                Defaults to True.

        Raises:
            IllumioApiException: if the response is unsuccessful (status code >399).

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        try:
            response = None  # avoid reference before assignment errors in case of cxn failure
            url = self._build_url(endpoint, include_org)
            self._encode_body(kwargs)
            response = self._session.request(method, url, **kwargs)
            response.raise_for_status()
            return response
        except Exception as e:
            message = str(e)
            # Response objects are falsy if the request failed so do a null check
            if response is not None and response.headers.get('Content-Type', '') == 'application/json':
                message = self._get_error_message_from_response(response)
            raise IllumioApiException(message) from e

    def _build_url(self, endpoint: str, include_org: bool):
        endpoint = endpoint.lstrip('/').replace('//', '/')
        if include_org and not endpoint.startswith('orgs/'):
            endpoint = 'orgs/{}/{}'.format(self.org_id, endpoint)
        return '{}/{}'.format(self.base_url, endpoint)

    def _encode_body(self, kwargs):
        """Encodes request body data to JSON."""
        body = kwargs.pop('data', None)
        if 'json' in kwargs:
            # json overrides data if both are provided
            body = kwargs.pop('json')
        if body is not None:
            kwargs['json'] = json.loads(self._encoder.encode(body))

    def _get_error_message_from_response(self, response: Response) -> str:
        message = "API call returned error code {}. Errors:".format(response.status_code)
        for error in response.json():
            if error and 'token' in error and 'message' in error:
                message += '\n{}: {}'.format(error['token'], error['message'])
            elif error and 'error' in error:
                message += '\n{}'.format(error['error'])
        return message

    def get(self, endpoint: str, **kwargs) -> Response:
        """Makes a GET call to a given PCE endpoint.

        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        return self._request('GET', endpoint, **kwargs)

    def post(self, endpoint: str, **kwargs) -> Response:
        """Makes a POST call to a given PCE endpoint.

        Appends 'Content-Type: application/json' to the request headers by default.
        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        headers = kwargs.get('headers', {})
        kwargs['headers'] = {**headers, **{'Content-Type': 'application/json'}}
        return self._request('POST', endpoint, **kwargs)

    def put(self, endpoint: str, **kwargs) -> Response:
        """Makes a PUT call to a given PCE endpoint.

        Appends 'Content-Type: application/json' to the request headers by default.
        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        headers = kwargs.get('headers', {})
        kwargs['headers'] = {**headers, **{'Content-Type': 'application/json'}}
        return self._request('PUT', endpoint, **kwargs)

    def delete(self, endpoint: str, **kwargs) -> Response:
        """Makes a DELETE call to a given PCE endpoint.

        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        return self._request('DELETE', endpoint, **kwargs)

    def get_collection(self, endpoint: str, **kwargs) -> Response:
        """Uses the PCE's asynchronous job logic to retrieve a collection of objects.

        NOTE: for large collections (surpassing 10,000 objects), this function will be extremely
        slow - it is recommended that callers use query filters or the max_results parameter to limit
        the number of results in the collection.

        Args:
            endpoint (str): the PCE endpoint to call.

        Raises:
            IllumioApiException: if an error message (>399) status code is returned.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        try:
            headers = kwargs.get('headers', {})
            kwargs['headers'] = {**headers, **{'Prefer': 'respond-async'}}
            response = self.get(endpoint, **kwargs)
            response.raise_for_status()
            location = response.headers['Location']
            retry_after = int(response.headers['Retry-After'])

            collection_href = self._async_poll(location, retry_after)

            response = self.get(collection_href, include_org=False)
            response.raise_for_status()
            return response
        except Exception as e:
            raise IllumioApiException from e

    def _async_poll(self, job_location: str, retry_time: Union[int, float] = 1.) -> str:
        """Polls the PCE for an async job's status until it completes or times out.

        The poll-wait loop uses the Retry-After time from the job submission  request
        or a default of 1 second, with a 1.5x backoff.

        Args:
            job_location (str): URL of the job to poll.
            retry_time (Union[int, float], optional): base Retry-After time. Defaults to 1 second.

        Raises:
            Exception: if the job returns a 'failed' status.

        Returns:
            str: the HREF path of the completed collection document.
        """
        while True:
            time.sleep(retry_time)
            retry_time *= 1.5  # slight backoff to avoid spamming the PCE for long-running jobs
            response = self.get(job_location)
            response.raise_for_status()
            poll_result = response.json()
            poll_status = poll_result['status']

            if poll_status == 'failed':
                raise Exception('Async collection job failed: ' + poll_result['result']['message'])
            elif poll_status == 'completed':
                # traffic flow async jobs
                collection_href = poll_result['result']
                break
            elif poll_status == 'done':
                # policy object collection jobs
                collection_href = poll_result['result']['href']
                break
        return collection_href

    def check_connection(self, **kwargs) -> bool:
        """Makes a GET request to the PCE /health endpoint.

        Additional keyword arguments are passed to the requests call.

        Returns:
            bool: True if the call is successful, otherwise False.
        """
        try:
            self.get('/health', include_org=False, **kwargs)
            return True
        except IllumioApiException:
            return False

    class _PCEObjectAPI:
        """Generic API for registered PCE objects.

        Each registered API exposes CRUD operation functions through this common interface.
        """
        def __init__(self, pce: 'PolicyComputeEngine', api_data: object) -> None:
            self.name = api_data.name
            self.endpoint = api_data.endpoint
            self.object_cls = api_data.object_class
            self.is_sec_policy = api_data.is_sec_policy
            self.is_global = api_data.is_global
            self.pce = pce

        def _build_endpoint(self, policy_version: str, parent: Any) -> str:
            """Builds the PCE request endpoint."""
            endpoint = self.endpoint

            if parent:  # e.g. /sec_policy/active/rulesets/1/sec_rules
                parent_draft_href = convert_active_href_to_draft(href_from(parent))
                endpoint = '{}/{}'.format(parent_draft_href, endpoint)
            else:  # mutually exclusive as the parent HREF will have the sec_policy and orgs prefix already
                if self.is_sec_policy:
                    if policy_version not in [ACTIVE, DRAFT]:
                        raise IllumioApiException("Invalid policy_version passed to get: {}".format(policy_version))
                    endpoint = '/sec_policy/{}/{}'.format(policy_version, endpoint)

                if not self.is_global:
                    endpoint = '/orgs/{}/{}'.format(self.pce.org_id, endpoint)
            return endpoint.replace('//', '/')

        def get_by_reference(self, reference: Union[str, Reference, dict], **kwargs) -> IllumioObject:
            """Retrieves an object from the PCE using its HREF.

            Usage:
                >>> ip_list = pce.ip_lists.get_by_reference('/orgs/1/sec_policy/active/ip_lists/1')
                >>> ip_list
                IPList(
                    name='Any (0.0.0.0/0 and ::/0)',
                    ...
                )

            Args:
                href (str): the HREF of the object to fetch.

            Returns:
                IllumioObject: the object json, decoded to its IllumioObject equivalent.
            """
            response = self.pce.get(href_from(reference), include_org=False, **kwargs)
            return self.object_cls.from_json(response.json())

        def get(self, policy_version: str = DRAFT, parent: Union[str, Reference, dict] = None, **kwargs) -> List[IllumioObject]:
            """Retrieves objects from the PCE based on the given parameters.

            Keyword arguments to this function are passed to the `requests.get` call.
            See https://docs.illumio.com/core/21.5/API-Reference/index.html
            for details on filter parameters for collection queries.

            Usage:
                >>> virtual_services = pce.virtual_services.get(
                ...     policy_version='active',
                ...     params={
                ...         'name': 'VS-'
                ...     }
                ... )
                >>> virtual_services
                [
                    VirtualService(
                        href='/orgs/1/sec_policy/active/virtual_services/9177c75f-7b21-4bf0-8c16-2c47c1ca3252',
                        name='VS-LAB-SERVICES'
                        ...
                    ),
                    ...
                ]

            Args:
                policy_version (str, optional): if fetching security policy objects, specifies
                    whether to fetch 'draft' or 'active' objects. Defaults to 'draft'.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing RuleSets.

            Returns:
                List[IllumioObject]: the returned list of decoded objects.
            """
            endpoint = self._build_endpoint(policy_version, parent)
            response = self.pce.get(endpoint, include_org=False, **kwargs)
            return [self.object_cls.from_json(o) for o in response.json()]

        def get_all(self, policy_version: str = DRAFT, parent: Union[str, Reference, dict] = None, **kwargs) -> List[IllumioObject]:
            """Retrieves all objects of a given type from the PCE.

            This function makes two requests, using the `X-Total-Count` header
            in the response to set the `max_results` parameter on the follow-up
            request.

            Args:
                policy_version (str, optional): if fetching security policy objects, specifies
                    whether to fetch 'draft' or 'active' objects. Defaults to 'draft'.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing RuleSets.

            Returns:
                List[IllumioObject]: the returned list of decoded objects.
            """
            params = kwargs.get('params', {})
            endpoint = self._build_endpoint(policy_version, parent)

            if 'max_results' not in params:
                kwargs['params'] = {**params, **{'max_results': 0}}
                response = self.pce.get(endpoint, **kwargs)
                if len(response.json()) > 0:  # for endpoints that don't support max_results
                    return [self.object_cls.from_json(o) for o in response.json()]
                filtered_object_count = response.headers['X-Total-Count']
                kwargs['params'] = {**params, **{'max_results': int(filtered_object_count)}}

            response = self.pce.get(endpoint, include_org=False, **kwargs)
            return [self.object_cls.from_json(o) for o in response.json()]

        def get_async(self, policy_version: str = DRAFT, parent: Union[str, Reference, dict] = None, **kwargs) -> List[IllumioObject]:
            """Retrieves objects asynchronously from the PCE based on the given parameters.

            Args:
                policy_version (str, optional): if fetching security policy objects, specifies
                    whether to fetch 'draft' or 'active' objects. Defaults to 'draft'.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing Rule Sets.

            Returns:
                List[IllumioObject]: the returned list of decoded objects.
            """
            endpoint = self._build_endpoint(policy_version, parent)
            response = self.pce.get_collection(endpoint, **kwargs)
            return [self.object_cls.from_json(o) for o in response.json()]

        def create(self, body: Any, parent: Union[str, Reference, dict] = None, **kwargs) -> IllumioObject:
            """Creates an object in the PCE.

            See https://docs.illumio.com/core/21.5/API-Reference/index.html
            for details on POST body parameters when creating objects.

            Usage:
                >>> from illumio.policyobjects import Label
                >>> label = Label(key='role', value='R-DB')
                >>> label = pce.labels.create(label)
                >>> label
                Label(
                    'href': '/orgs/1/labels/14',
                    'key': 'role',
                    'value': 'R-DB
                )

            Args:
                body (Any): the parameters for the newly created object.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing RuleSets.

            Returns:
                IllumioObject: the created object.
            """
            kwargs['json'] = body
            endpoint = self._build_endpoint(DRAFT, parent)
            response = self.pce.post(endpoint, include_org=False, **kwargs)
            return self._parse_response_body(response.json())

        def _parse_response_body(self, json_response):
            # XXX: workaround for Service Bindings. Multiple bindings
            #   can be created in the same POST, so we need to accommodate
            #   this case by checking the response body type
            if type(json_response) is list:
                results = {self.name: [], 'errors': []}
                for o in json_response:
                    if 'href' in o:
                        results[self.name].append(self.object_cls.from_json(o))
                    else:
                        results['errors'].append(o)
                return results
            return self.object_cls.from_json(json_response)

        def update(self, reference: Union[str, Reference, dict], body: Any, **kwargs) -> None:
            """Updates an object in the PCE.

            Successful PUT requests return a 204 No Content response.

            Usage:
                >>> from illumio.workloads import PairingProfile
                >>> pairing_profiles = pce.pairing_profile.get(
                ...     params={'name': 'PP-DATABASE', 'max_results': 1}
                ... )
                >>> existing_profile = pairing_profiles[0]
                >>> update = PairingProfile(
                ...     name='PP-DATABASE-VENS',
                ...     enabled=False  # disable this profile
                ... )
                >>> pce.pairing_profile.update(existing_profile['href'], update)

            Args:
                reference (Union[str, Reference, dict]): the HREF of the pairing profile to update.
                body (Any): the update data.
            """
            kwargs['json'] = body
            self.pce.put(href_from(reference), include_org=False, **kwargs)

        def delete(self, reference: Union[str, Reference, dict], **kwargs) -> None:
            """Deletes an object in the PCE.

            Successful DELETE requests return a 204 No Content response.

            Args:
                reference (Union[str, Reference, dict]): the HREF of the object to delete.
            """
            self.pce.delete(href_from(reference), include_org=False, **kwargs)

        def _bulk_change(self, objects: List[IllumioObject], method: str, success_status: str, **kwargs) -> List[dict]:
            results = []
            while objects:
                kwargs['json'] = objects[:BULK_CHANGE_LIMIT]
                objects = objects[BULK_CHANGE_LIMIT:]
                endpoint = self._build_endpoint(DRAFT, None)
                response = self.pce.put('{}/{}'.format(endpoint, method), include_org=False, **kwargs)
                for result in response.json():
                    errors = result.get('errors', [])
                    if success_status and result['status'] != success_status:
                        errors.append({'token': result['token'], 'message': result['message']})
                    results.append({'href': result['href'], 'errors': errors})
            return results

        def bulk_create(self, objects_to_create: List[IllumioObject], **kwargs) -> List[dict]:
            """Creates a set of objects in the PCE.

            NOTE: Bulk creation can currently only be applied for Security Principals,
                Virtual Services and Workloads.

            Args:
                objects_to_create (List[IllumioObject]): list of objects to update.

            Returns:
                List[dict]: a list containing HREFs of created objects
                    as well as any errors returned from the PCE.
                    Has the following form:

                    [
                        {
                            'href': {object_href},
                            'errors': [
                                {
                                    'token': {error_type},
                                    'message': {error_message}
                                }
                            ]
                        }
                    ]
            """
            return self._bulk_change(objects_to_create, method='bulk_create', success_status='created', **kwargs)

        def bulk_update(self, objects_to_update: List[IllumioObject], **kwargs) -> List[dict]:
            """Updates a set of objects in the PCE.

            NOTE: Bulk updates can currently only be applied for Virtual Services and Workloads.

            Args:
                objects_to_update (List[IllumioObject]): list of objects to update.

            Returns:
                List[dict]: a list containing HREFs of updated objects
                    as well as any errors returned from the PCE.
                    Has the following form:

                    [
                        {
                            'href': {object_href},
                            'errors': [
                                {
                                    'token': {error_type},
                                    'message': {error_message}
                                }
                            ]
                        }
                    ]
            """
            return self._bulk_change(objects_to_update, method='bulk_update', success_status='updated', **kwargs)

        def bulk_delete(self, refs: List[Union[str, Reference, dict]], **kwargs) -> List[dict]:
            """Deletes a set of objects in the PCE.

            NOTE: Bulk updates can currently only be applied for Workloads.

            Args:
                hrefs (List[Union[str, Reference, dict]]): list of references to objects to delete.

            Returns:
                List[dict]: a list containing any errors that occurred during
                    the bulk operation. Has the following form:

                    [
                        {
                            'href': {object_href},
                            'errors': [
                                {
                                    'token': {error_type},
                                    'message': {error_message}
                                }
                            ]
                        }
                    ]
            """
            objects_to_delete = [Reference(href=href_from(reference)) for reference in refs]
            return self._bulk_change(objects_to_delete, method='bulk_delete', success_status=None, **kwargs)

    def __getattr__(self, name: str) -> _PCEObjectAPI:
        """Instantiates a generic API for registered PCE objects.

        Inspired by the Zabbix API: https://pypi.org/project/zabbix-api/
        """
        if name in self._apis:
            return self._apis[name]
        if name not in PCE_APIS:
            raise AttributeError("No such PCE API object: {}".format(name))
        api = self._PCEObjectAPI(pce=self, api_data=PCE_APIS[name])
        self._apis[name] = api
        return api

    def get_default_ip_list(self, **kwargs) -> IPList:
        """Retrieves the "Any (0.0.0.0/0 and ::/0)" default global IP list.

        Returns:
            IPList: decoded object representing the default global IP list.
        """
        params = kwargs.get('params', {})
        # retrieve by name as each org will use a different ID
        kwargs['params'] = {**params, **{'name': ANY_IP_LIST_NAME}}
        response = self.get('/sec_policy/active/ip_lists', **kwargs)
        return IPList.from_json(response.json()[0])

    def generate_pairing_key(self, pairing_profile_href: str, **kwargs) -> str:
        """Generates a pairing key using a pairing profile.

        Args:
            pairing_profile_href (str): the HREF of the pairing profile to use.

        Returns:
            str: the pairing key value.
        """
        kwargs['json'] = {}
        response = self.post('{}/pairing_key'.format(pairing_profile_href), include_org=False, **kwargs)
        return response.json().get('activation_code')

    @deprecated(deprecated_in='1.0.0')
    def get_traffic_flows(self, traffic_query: TrafficQuery, **kwargs) -> List[TrafficFlow]:
        """DEPRECATED (v1.0.0). Use `get_traffic_flows_async` instead.

        Retrieves Explorer traffic flows using the provided query.

        NOTE: this function is deprecated in the Illumio REST API, and is
        only provided for compatibility. The Illumio Explorer REST API
        documentation recommends using the async traffic flow query instead,
        provided here as `PolicyComputeEngine.get_traffic_flows_async`.

        See https://docs.illumio.com/core/21.5/Content/Guides/rest-api/visualization/explorer.htm#TrafficAnalysisQueries
        for details on traffic query parameters.

        Args:
            traffic_query (TrafficQuery): `TrafficQuery` object representing
                the query parameters.

        Returns:
            List[TrafficFlow]: list of `TrafficFlow` objects found using the
                provided query.
        """
        kwargs['json'] = traffic_query
        response = self.post('/traffic_flows/traffic_analysis_queries', **kwargs)
        return [TrafficFlow.from_json(flow) for flow in response.json()]

    def get_traffic_flows_async(self, query_name: str, traffic_query: TrafficQuery, **kwargs) -> List[TrafficFlow]:
        """Retrieves Explorer traffic flows using the provided query.

        See https://docs.illumio.com/core/21.5/Content/Guides/rest-api/visualization/explorer.htm#AsynchronousQueriesforTrafficFlows
        for details on async traffic query parameters.

        Usage:
            >>> from illumio.explorer import TrafficQuery
            >>> traffic_query = TrafficQuery.build(
            ...     start_date="2022-02-01T00:00:00Z",
            ...     end_date="2022-03-01T00:00:00Z",
            ...     include_services=[
            ...         {'port': 3389, 'proto': 'tcp'}
            ...     ],
            ...     policy_decisions=['potentially_blocked', 'unknown']
            ... )
            >>> traffic_query
            TrafficQuery(
                start_date='2022-02-01T00:00:00Z',
                end_date='2022-03-01T00:00:00Z',
                sources=TrafficQueryFilterBlock(
                    include=[],
                    exclude=[]
                ),
                destinations=TrafficQueryFilterBlock(
                    include=[],
                    exclude=[]
                ),
                services=TrafficQueryServiceBlock(
                    include=[
                        ServicePort(
                            port=3389,
                            proto=6,
                            ...
                        )
                    ],
                    exclude=[]
                ),
                policy_decisions=[
                    'potentially_blocked',
                    'unknown'
                ],
                ...
            )
            >>> traffic_flows = pce.get_traffic_flows_async(
            ...     query_name='rdp-traffic-feb-22',
            ...     traffic_query=traffic_query
            ... )
            >>> traffic_flows
            [
                TrafficFlow(
                    src=TrafficNode(
                        ip='10.0.9.14',
                        "ip_lists": [
                            {
                                "name": "IPL-INTERNAL",
                                "href": "/orgs/1/sec_policy/active/ip_lists/5",
                                ...
                            }
                        ],
                        ...
                    ),
                    dst=TrafficNode(
                        ip='10.0.6.63',
                        workload=Workload(
                            href='/orgs/1/workloads/6567900a-b49f-43cc-93a7-c892da39aad1',
                            name='WIN-JUMPBOX',
                            ...
                        ),
                        ...
                    ),
                    service=ServicePort(port=3389, proto=6),
                    num_connections=44,
                    state='closed',
                    timestamp_range=TimestampRange(
                        first_detected='2022-02-19T09:50:17Z',
                        last_detected='2022-02-19T10:12:36Z'
                    ),
                    policy_decision='potentially_blocked',
                    flow_direction='inbound',
                    ...
                ),
                ...
            ]

        Args:
            query_name (str): name for the async query job.
            traffic_query (TrafficQuery): `TrafficQuery` object representing
                the query parameters.

        Raises:
            IllumioApiException: if there is an error retrieving the async job
                results.

        Returns:
            List[TrafficFlow]: list of `TrafficFlow` objects found using the
                provided query.
        """
        try:
            traffic_query.query_name = query_name
            kwargs['json'] = traffic_query
            headers = kwargs.get('headers', {})
            kwargs['headers'] = {**headers, **{'Content-Type': 'application/json', 'Prefer': 'respond-async'}}
            response = self.post('/traffic_flows/async_queries', **kwargs)
            response.raise_for_status()
            query_status = response.json()
            location = query_status['href']

            collection_href = self._async_poll(location)

            response = self.get(collection_href, include_org=False)
            response.raise_for_status()
            return [TrafficFlow.from_json(flow) for flow in response.json()]
        except Exception as e:
            raise IllumioApiException from e

    def provision_policy_changes(self, change_description: str, hrefs: List[str], **kwargs) -> PolicyVersion:
        """Provisions policy changes for draft objects with the given HREFs.

        Usage:
            >>> from illumio.rules import RuleSet
            >>> rule_set = pce.rule_sets.create(
            ...     RuleSet(name='RS-RINGFENCE')
            ... )
            >>> changeset = pce.provision_policy_changes(
            ...     change_description='Provision ring-fence rule set',
            ...     hrefs=[rule_set.href]
            ... )
            >>> changeset
            PolicyVersion(
                href='/orgs/1/sec_policy/110',
                commit_message='Provision ring-fence rule set',
                version=110,
                workloads_affected=0,
                object_counts=PolicyObjectCounts(
                    label_groups=17,
                    services=8,
                    ...
                )
            )

        Args:
            change_description (str): the policy change description.
            hrefs (List[str]): the HREFs of the draft policy objects to provision.

        Raises:
            IllumioException: if an invalid HREF is provided.

        Returns:
            PolicyVersion: the decoded policy version object including the changeset.
        """
        policy_changeset = PolicyChangeset.build(hrefs)
        kwargs['json'] = {'update_description': change_description, 'change_subset': policy_changeset}
        response = self.post('/sec_policy', **kwargs)
        return PolicyVersion.from_json(response.json())


__all__ = ['PolicyComputeEngine']
